from pathlib import Path
from unittest.mock import MagicMock, patch

from pydantic import SecretStr
from pytest import raises

from monopoly.pdf import (
    BadPasswordFormatError,
    MissingPasswordError,
    PdfDocument,
    WrongPasswordError,
)

fixture_directory = Path(__file__).parent / "fixtures"


def test_document_initialization_with_both_raises_error():
    file_path = Path("src/monopoly/examples/example_statement.pdf")
    with raises(RuntimeError, match="Only one of `file_path` or `file_bytes` should be passed"):
        PdfDocument(file_path=file_path, file_bytes=b"123")


def test_document_initialization_with_neither_raises_error():
    with raises(RuntimeError, match="Either `file_path` or `file_bytes` must be passed"):
        PdfDocument()


def test_can_open_file_stream():
    with open(fixture_directory / "4_pages_blank.pdf", "rb") as file:
        pdf_document = PdfDocument(file_bytes=file.read())
        assert len(pdf_document) == 4


def test_can_open_protected(pdf_document: PdfDocument):
    mock_pdf_passwords_instance = MagicMock()
    mock_pdf_passwords_instance.pdf_passwords = [SecretStr("foobar123")]

    with patch("monopoly.pdf.PdfPasswords", return_value=mock_pdf_passwords_instance):
        pdf_document = PdfDocument(passwords=None, file_path=fixture_directory / "protected.pdf")
        pdf_document.unlock_document()


def test_wrong_password_raises_error():
    with raises(WrongPasswordError, match="Could not open"):
        pdf_document: PdfDocument = PdfDocument(
            passwords=[SecretStr("wrongpw_123")],
            file_path=fixture_directory / "protected.pdf",
        )
        pdf_document.unlock_document()


def test_override_password():
    pdf_document: PdfDocument = PdfDocument(
        passwords=[SecretStr("foobar123")],
        file_path=fixture_directory / "protected.pdf",
    )
    pdf_document.unlock_document()
    assert not pdf_document.is_encrypted


def test_missing_password_raises_error():
    mock_pdf_passwords_instance = MagicMock()
    mock_pdf_passwords_instance.pdf_passwords = []

    with raises(MissingPasswordError):
        with patch("monopoly.pdf.PdfPasswords", return_value=mock_pdf_passwords_instance):
            pdf_document = PdfDocument(passwords=None, file_path=fixture_directory / "protected.pdf")
            pdf_document.unlock_document()


def test_null_password_raises_error():
    mock_pdf_passwords_instance = MagicMock()
    mock_pdf_passwords_instance.pdf_passwords = [SecretStr("")]

    with raises(MissingPasswordError, match="is empty"):
        with patch("monopoly.pdf.PdfPasswords", return_value=mock_pdf_passwords_instance):
            pdf_document = PdfDocument(passwords=None, file_path=fixture_directory / "protected.pdf")
            pdf_document.unlock_document()


def test_invalid_password_type_raises_error():
    mock_pdf_passwords_instance = MagicMock()
    mock_pdf_passwords_instance.pdf_passwords = "not a list"

    with raises(BadPasswordFormatError, match="should be stored in a list"):
        with patch("monopoly.pdf.PdfPasswords", return_value=mock_pdf_passwords_instance):
            pdf_document = PdfDocument(passwords=None, file_path=fixture_directory / "protected.pdf")
            pdf_document.unlock_document()


def test_plain_text_passwords_raises_error(pdf_document: PdfDocument):
    mock_pdf_passwords_instance = MagicMock()
    mock_pdf_passwords_instance.pdf_passwords = ["insecure"]

    with raises(BadPasswordFormatError, match="should be stored as SecretStr"):
        with patch("monopoly.pdf.PdfPasswords", return_value=mock_pdf_passwords_instance):
            pdf_document = PdfDocument(passwords=None, file_path=fixture_directory / "protected.pdf")
            pdf_document.unlock_document()



from dataclasses import dataclass, field
from pathlib import Path

import click


@dataclass
class RunConfig:
    output_directory: Path | None = None
    pprint: bool = False
    safety_check: bool = True
    single_process: bool = False
    use_ocr: bool = False
    verbose: bool = False
    preserve_filename: bool = False


@dataclass
class TqdmSettings:
    """Configuration for a tqdm progress bar."""

    total: int
    desc: str = "Processing statements"
    leave: bool = False
    delay: float = 0.2
    ncols: int = 80
    bar_format: str = "{l_bar}{bar}| {n_fmt}/{total_fmt}"


@dataclass
class Result:
    """Stores information about processed bank statement."""

    source_file_name: str
    target_file_name: str | None = None
    error_info: dict[str, str] = field(default_factory=dict)


@dataclass
class Report:
    """A helper class for parsing and displaying processed bank statements."""

    results: list[Result]

    @property
    def processed_results(self) -> list[Result]:
        return sorted(
            [r for r in self.results if not r.error_info],
            key=lambda x: x.source_file_name,
        )

    @property
    def errored_results(self) -> list[Result]:
        return sorted([r for r in self.results if r.error_info], key=lambda x: x.source_file_name)

    @property
    def number_processed(self) -> int:
        return len(self.processed_results)

    @property
    def number_errored(self) -> int:
        return len(self.errored_results)

    def display_report(self, *_, verbose=False) -> None:
        """Parse all results, displaying the number of successfully processed statements and any errors."""
        for res in self.processed_results:
            click.echo(f"{res.source_file_name} -> {res.target_file_name}")

        if self.number_errored > 0:
            error_msg = f"{self.number_errored} statement(s) had errors while processing"
            click.echo(click.style(error_msg, fg="red", bold=True))
        if self.number_processed > 0:
            changed_msg = f"{self.number_processed} statement(s) processed"
            click.echo(click.style(changed_msg, bold=True))

        for res in self.errored_results:
            error_msg = res.error_info["message"]
            if verbose:
                error_msg = res.error_info["traceback"]
            click.echo(
                click.style(
                    f"{res.source_file_name} -- {error_msg}",
                    fg="red",
                )
            )



import logging

from .amex import Amex
from .bank_of_america import BankOfAmerica
from .base import BankBase
from .bmo import BankOfMontreal
from .canadian_tire import CanadianTire
from .capitalone import CapitalOneCanada
from .chase import Chase
from .cibc import CIBC
from .citibank import Citibank
from .dbs import Dbs
from .detector import BankDetector
from .example_bank import ExampleBank
from .hsbc import Hsbc
from .maybank import Maybank
from .ocbc import Ocbc
from .rbc import RoyalBankOfCanada
from .scotiabank import Scotiabank
from .standard_chartered import StandardChartered
from .td_canada_trust import TDCanadaTrust
from .trust import Trust
from .uob import Uob
from .zkb import ZurcherKantonalBank

banks: list[type["BankBase"]] = [
    Amex,
    BankOfAmerica,
    BankOfMontreal,
    CanadianTire,
    CapitalOneCanada,
    Chase,
    CIBC,
    Citibank,
    Dbs,
    ExampleBank,
    Hsbc,
    Maybank,
    Ocbc,
    Scotiabank,
    StandardChartered,
    Uob,
    ZurcherKantonalBank,
    Trust,
    TDCanadaTrust,
    RoyalBankOfCanada,
]

logger = logging.getLogger(__name__)

__all__ = ["BankDetector", "BankBase"] + [bank.__name__ for bank in banks]



"""Store statement-related enums and constants."""

from enum import auto

from strenum import StrEnum

from monopoly.enums import AutoEnum, RegexEnum

from .date import ISO8601, DateFormats


class EntryType(AutoEnum):
    CREDIT = auto()
    DEBIT = auto()


class BankNames(AutoEnum):
    AMEX = auto()
    BANK_OF_AMERICA = auto()
    CAPITAL_ONE = auto()
    CANADIAN_TIRE = auto()
    CHASE = auto()
    CIBC = auto()
    CITIBANK = auto()
    DBS = auto()
    HSBC = auto()
    MAYBANK = auto()
    OCBC = auto()
    RBC = auto()
    SCOTIABANK = auto()
    STANDARD_CHARTERED = auto()
    UOB = auto()
    ZKB = auto()
    TRUST = auto()
    TDCT = auto()
    BMO = auto()


class InternalBankNames(AutoEnum):
    EXAMPLE = auto()
    GENERIC = auto()


class Columns(AutoEnum):
    AMOUNT = auto()
    DATE = auto()
    DESCRIPTION = auto()
    POLARITY = auto()
    TRANSACTION_DATE = auto()


class SharedPatterns(StrEnum):
    """
    Attempt to keep patterns DRY.

    AMOUNT matches the following patterns:
    1,123.12 | 123.12 | (123.12) | ( 123.12) | 123.12 CR | -1,123.12.

    AMOUNT_EXTENDED is generally used for credit statements and to
    find statement balances, while AMOUNT_EXTENDED_WITHOUT_EOL is used
    for debit statements, since debit statements have the amount + balances
    on a single line, and we only want the amount.
    """

    COMMA_FORMAT = r"\d{1,3}(,\d{3})*\.\d*"
    ENCLOSED_COMMA_FORMAT = rf"\({COMMA_FORMAT}\s{{0,1}}\))"
    OPTIONAL_NEGATIVE_SYMBOL = r"(?:-)?"
    POLARITY = r"(?P<polarity>CR\b|DR\b|DB\b|\+|\-)?\s*"

    AMOUNT = rf"(?P<amount>{COMMA_FORMAT}|{ENCLOSED_COMMA_FORMAT}\s*"
    AMOUNT_EXTENDED_WITHOUT_EOL = AMOUNT + POLARITY
    AMOUNT_EXTENDED = AMOUNT_EXTENDED_WITHOUT_EOL + r"$"

    BALANCE = rf"(?P<balance>{COMMA_FORMAT})?$"
    DESCRIPTION = r"(?P<description>.*?)\s+"
    TRANSACTION_DATE_ABBREVIATED_ALL_CAPS = r"(?P<transaction_date>\d{2}\s[A-Z]{3})\s+"
    TRANSACTION_DATE_ABBREVIATED_PROPER_CASE = r"(?P<transaction_date>\d{2}\s[A-Z]{1}[a-z]{2})\s+"
    POSTING_DATE_ABBREVIATED_PROPER = r"(?P<posting_date>\d{2}\s[A-Z]{1}[a-z]{2})\s+"


class StatementBalancePatterns(RegexEnum):
    DBS = r"(?P<description>PREVIOUS BALANCE?)\s+" + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    CITIBANK = r"(?P<description>BALANCE PREVIOUS STATEMENT?)\s+" + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    HSBC = r"(?P<description>Previous Statement Balance?)\s+" + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    MAYBANK_MY = r"(?P<description>YOUR PREVIOUS STATEMENT BALANCE?)\s+" + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    MAYBANK_SG = (
        r"(?P<description>OUTSTANDING\s+BALANCE\s+BROUGHT\s+FORWARD?)\s+" + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    OCBC = r"(?P<description>LAST MONTH'S BALANCE?)\s+" + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    STANDARD_CHARTERED = (
        r"(?P<description>BALANCE FROM PREVIOUS STATEMENT?)\s+" + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    UOB = r"(?P<description>PREVIOUS BALANCE?)\s+" + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    TRUST = r"(?P<description>Previous balance?)\s+" + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    TDCT = r"(?P<description>PREVIOUS STATEMENT BALANCE?)\s+" + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    RBC = r"(?P<description>PREVIOUS STATEMENT BALANCE?)\s+" + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    CIBC = (
        r"(?P<description>Previous balance?)\s+"
        rf"(?P<amount>{SharedPatterns.OPTIONAL_NEGATIVE_SYMBOL}\$?{SharedPatterns.COMMA_FORMAT})"
    )


class CreditTransactionPatterns(RegexEnum):
    AMEX_PLATINUM = (
        rf"(?P<transaction_date>{ISO8601.DD_MM_YY})\s+" + SharedPatterns.DESCRIPTION + SharedPatterns.AMOUNT_EXTENDED
    )
    BANK_OF_AMERICA = (
        rf"(?P<transaction_date>{ISO8601.MM_DD})\s+"
        rf"(?P<posting_date>{ISO8601.MM_DD})\s+"
        + SharedPatterns.DESCRIPTION
        + r"(?P<reference_number>\d{4})?\s+"
        + r"(?P<account_number>\d{4})?\s+"
        + r"(?P<polarity>\-)?"
        + SharedPatterns.AMOUNT
    )
    BMO = (
        r"(?P<transaction_date>[A-Z][a-z]{2,3}\.\s+\d{1,2})\s+"
        r"(?P<posting_date>[A-Z][a-z]{2,3}\.\s+\d{1,2})\s+"
        r"(?P<description>.+?)\s+"
        r"(?P<amount>\d{1,3}(?:,\d{3})*\.\d{2})(\s+)?"
        r"(?P<polarity>CR)?$"
    )
    DBS = rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+" + SharedPatterns.DESCRIPTION + SharedPatterns.AMOUNT_EXTENDED
    CANADIAN_TIRE = (
        r"^\s*"
        r"(?P<transaction_date>[A-Z][a-z]{2}\s\d{2})\s+"
        r"(?P<posting_date>[A-Z][a-z]{2}\s\d{2})\s+"
        r"(?!\s*\d+\b)"
        r"(?P<description>.+?)\s{2,}"  # NOTE: no way to not parse trailing text in line as description?
        r"(?P<polarity>-)"
        r"?(?P<amount>\d{1,3}(?:,\d{3})*\.\d{2})"
    )
    CAPITAL_ONE = (
        r"^\s*"
        rf"(?P<transaction_date>\b({DateFormats.MMM}\s+{DateFormats.D}))\s+"
        rf"(?P<posting_date>\b({DateFormats.MMM}\s+{DateFormats.D}))\s+"
        f"{SharedPatterns.DESCRIPTION}"
        rf"(?P<polarity>-)?\s*\$"
        rf"(?P<amount>{SharedPatterns.COMMA_FORMAT})\s*$"
    )
    CHASE = (
        rf"(?P<transaction_date>{ISO8601.MM_DD})\s+"
        + SharedPatterns.DESCRIPTION
        + r"(?P<polarity>\-)?"
        + r"(?P<amount>(\d{1,3}(,\d{3})*|\d*)\.\d+)$"
    )
    CIBC = (
        rf"(?P<transaction_date>\b({DateFormats.MMM}[-\s]{DateFormats.DD}))\s+"
        rf"(?P<posting_date>\b({DateFormats.MMM}[-\s]{DateFormats.DD}))\s+"
        r"(?:Ý\s*)?(?P<description>.+?)\s{2,}"
        r"(?P<catagory>.*?)\s+"
        # transaction dr/cr with format -$999,000.00
        rf"(?P<amount>{SharedPatterns.OPTIONAL_NEGATIVE_SYMBOL}\$?{SharedPatterns.COMMA_FORMAT}|{SharedPatterns.ENCLOSED_COMMA_FORMAT}\s*"
    )
    CITIBANK = (
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+" + SharedPatterns.DESCRIPTION + SharedPatterns.AMOUNT_EXTENDED
    )
    HSBC = (
        rf"(?P<posting_date>{ISO8601.DD_MMM_RELAXED})\s+"
        rf"(?P<transaction_date>{ISO8601.DD_MMM_RELAXED})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    MAYBANK_SG = (
        r"(?P<posting_date>\b[A-Z]?\d{1,2}(JAN|FEB|MAR|APR|MAY|JUN|JUL|AUG|SEP|OCT|NOV|DEC)\b)\s+"
        r"(?P<transaction_date>\b[A-Z]?\d{1,2}(JAN|FEB|MAR|APR|MAY|JUN|JUL|AUG|SEP|OCT|NOV|DEC)\b)\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    MAYBANK_MY = (
        rf"(?P<posting_date>{ISO8601.DD_MM})\s+"
        rf"(?P<transaction_date>{ISO8601.DD_MM})\s+" + SharedPatterns.DESCRIPTION + SharedPatterns.AMOUNT_EXTENDED
    )
    OCBC = r"(?P<transaction_date>\d+/\d+)\s+" + SharedPatterns.DESCRIPTION + SharedPatterns.AMOUNT_EXTENDED
    RBC = (
        rf"(?P<transaction_date>\b({DateFormats.MMM}[-\s]{DateFormats.DD}))\s+"
        rf"(?P<posting_date>\b({DateFormats.MMM}[-\s]{DateFormats.DD}))\s+"
        + SharedPatterns.DESCRIPTION
        # transaction dr/cr with format -$999,000.00
        + r"(?P<polarity>\-)?"
        + rf"(?P<amount>\$?{SharedPatterns.COMMA_FORMAT}|{SharedPatterns.ENCLOSED_COMMA_FORMAT}\s*"
    )
    SCOTIABANK = (
        rf"(?P<transaction_date>\b({DateFormats.MMM}[\/\-\s.]{DateFormats.D}))\s+"
        rf"(?P<posting_date>\b({DateFormats.MMM}[\/\-\s.]{DateFormats.D}))\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL  # "credits" are denoted at the end of the amount, i.e PMYT 155.96-
    )
    STANDARD_CHARTERED = (
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+"
        rf"(?P<posting_date>{ISO8601.DD_MMM})\s+"
        + SharedPatterns.DESCRIPTION
        + r"(?:(?P<transaction_ref>Transaction\sRef\s\d+)?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED
    )
    UOB = (
        rf"(?P<posting_date>{ISO8601.DD_MMM})\s+"
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+" + SharedPatterns.DESCRIPTION + SharedPatterns.AMOUNT_EXTENDED
    )
    TDCT = (
        rf"(?P<transaction_date>\b({DateFormats.MMM}[-\s]{DateFormats.D}))\s+"
        rf"(?P<posting_date>\b({DateFormats.MMM}[-\s]{DateFormats.D}))\s+"
        + SharedPatterns.DESCRIPTION
        # transaction dr/cr with format -$999,000.00
        + rf"(?P<amount>{SharedPatterns.OPTIONAL_NEGATIVE_SYMBOL}\$?{SharedPatterns.COMMA_FORMAT})\s*"
    )
    TRUST = (
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+"
        r"(?P<description>(?:(?!Total outstanding balance).)*?)"
        r"(?P<polarity>\+)?"
        f"{SharedPatterns.AMOUNT}"
        r"$"  # necessary to ignore FCY
    )


class DebitTransactionPatterns(RegexEnum):
    BANK_OF_AMERICA = (
        rf"(?P<transaction_date>{ISO8601.MM_DD_YY})\s+"
        + SharedPatterns.DESCRIPTION
        + r"(?P<polarity>\-)?"
        + SharedPatterns.AMOUNT
    )
    BMO = (
        rf"^(?!.*(?:Closing totals)).*?"
        rf"(?P<transaction_date>{ISO8601.MMM_DD})\s+"
        rf"{SharedPatterns.DESCRIPTION}\s+"
        rf"(?P<amount>{SharedPatterns.COMMA_FORMAT})\s+"
        rf"(?P<balance>{SharedPatterns.COMMA_FORMAT})$"
    )
    CIBC = (
        rf"\s*(?:(?P<transaction_date>{DateFormats.MMM}\s+{DateFormats.D})[-\s]+)?"
        r"(?P<description>(?!(Deposits)).+?)\s{2,}"
        rf"(?P<amount>{SharedPatterns.COMMA_FORMAT})"
        rf"[-\s]+(?P<balance>{SharedPatterns.OPTIONAL_NEGATIVE_SYMBOL}\$?{SharedPatterns.COMMA_FORMAT})"
    )
    DBS = (
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    DBS_POSB_CONSOLIDATED = (
        rf"(?P<transaction_date>{ISO8601.DD_MM_YYYY})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    MAYBANK_MY = (
        rf"(?P<transaction_date>{ISO8601.DD_MM_YY})\s+"
        + SharedPatterns.DESCRIPTION
        # remove *\s
        + SharedPatterns.AMOUNT[:-3]
        + r"(?P<polarity>\-|\+)\s+"
        + SharedPatterns.BALANCE
    )
    OCBC = (
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+"
        rf"(?P<posting_date>{ISO8601.DD_MMM})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
        + SharedPatterns.BALANCE
    )
    RBC = (
        r"^(?!.*(?:Opening Balance|Closing Balance))"
        r"(\s*)?"
        rf"(?P<transaction_date>{DateFormats.D}\s+{DateFormats.MMM})?\s+"  # i.e 7 May
        r"(?P<description>.+?)\s{2,}"
        rf"(?P<amount>{SharedPatterns.COMMA_FORMAT})"
        r"(?:\s{2,}(?P<balance>-?\s?\d{1,3}(?:,\d{3})*(?:\.\d{1,2})))?"  # edge case for literally "- 100.00"
        r"\s*$"
    )
    SCOTIABANK = (
        r"^(?!.*(?:Opening Balance)).*?"  # avoid matching opening balance as a "transaction"
        rf"(?P<transaction_date>{ISO8601.MMM_DD})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT
        + SharedPatterns.BALANCE
    )
    UOB = (
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
        + SharedPatterns.BALANCE
    )
    ZKB = (
        rf"(?P<transaction_date>{ISO8601.DD_MM_YYYY})\s+"
        + SharedPatterns.DESCRIPTION
        + r"(?P<amount>\d{1,3}(\'\d{3})*(\.\d+)?)\s+"
        + rf"(?P<value_date>{ISO8601.DD_MM_YYYY})\s+"
        + r"(?P<balance>\d{1,3}(\'\d{3})*(\.\d+)?)$"
    )
    TDCT = (
        SharedPatterns.DESCRIPTION
        + rf"(?P<amount>{SharedPatterns.COMMA_FORMAT})"
        + rf"[-\s]+(?P<transaction_date>\b({DateFormats.MMM}{DateFormats.DD}))"  # i.e MAY1 and MAY27
        + rf"([-\s]+(?P<balance>{SharedPatterns.COMMA_FORMAT}))?"  # balance is shown at end of each day
    )



import traceback
from collections.abc import Collection, Iterable
from concurrent.futures import ProcessPoolExecutor
from dataclasses import asdict
from functools import partial
from pathlib import Path

import click
from tabulate import tabulate
from tqdm import tqdm

from monopoly.cli.models import Report, Result, RunConfig, TqdmSettings
from monopoly.log import file_context, setup_logs, worker_log_setup


def process_statement(file: Path, config: RunConfig) -> Result | None:
    """Extract, transform, and load transactions from bank statements."""
    # The file_context should only be set and displayed when the verbose formatter is active,
    # but setting an additional gate here to adopt a slightly stricter security posture.
    if config.verbose:
        file_context.set(file.name)

    # Lazily importing here prevents from CLI from having a "slow" start
    from monopoly.banks import BankDetector, banks
    from monopoly.generic import GenericBank
    from monopoly.pdf import PdfDocument, PdfParser
    from monopoly.pipeline import Pipeline

    try:
        document = PdfDocument(file)
        document.unlock_document()

        if config.use_ocr:
            document = PdfParser.apply_ocr(document)

        analyzer = BankDetector(document)
        bank = analyzer.detect_bank(banks) or GenericBank
        parser = PdfParser(bank, document)
        pipeline = Pipeline(parser)

        statement = pipeline.extract(safety_check=config.safety_check)
        transactions = pipeline.transform(statement)

        if config.pprint:
            pprint_transactions(transactions, statement, file)
            # don't load to CSV if pprint
            return None

        # saves processed statements to the same directory by default
        if not config.output_directory:
            config.output_directory = file.parent

        output_file = pipeline.load(
            transactions,
            statement,
            config.output_directory,
            preserve_filename=config.preserve_filename,
        )
        return Result(file.name, output_file.name)

    # ruff: noqa: BLE001
    except Exception as err:
        error_info = {
            "message": f"{type(err).__name__}: {err!s}",
            "traceback": traceback.format_exc(),
        }
        return Result(file.name, error_info=error_info)


def pprint_transactions(transactions: list, statement, file: Path) -> None:
    """Print transactions in a markdown style tabular format."""
    click.echo(f"{file.name}")
    transactions_as_dict = [transaction.as_raw_dict() for transaction in transactions]
    headers = {col: col for col in statement.columns}
    click.echo(
        tabulate(
            transactions_as_dict,
            headers=headers,
            tablefmt="psql",
            numalign="right",
        )
    )
    click.echo()


def get_results(input_files: Collection[Path], config: RunConfig):
    tqdm_settings = asdict(TqdmSettings(len(input_files)))
    processor = partial(process_statement, config=config)

    if config.single_process or len(input_files) == 1:
        return [processor(file) for file in tqdm(input_files, **tqdm_settings)]

    initializer = partial(worker_log_setup, verbose=config.verbose)

    with ProcessPoolExecutor(initializer=initializer) as executor:
        return list(
            tqdm(
                executor.map(processor, input_files),
                **tqdm_settings,
            )
        )


def run(input_files: Collection[Path], config: RunConfig):
    """
    Process a collection of input files concurrently.

    If any statements are processed successfully or encounter errors, a Report object
    is created and its display_report() method is called to provide a summary
    of the processing outcomes.
    """
    results = get_results(input_files, config)

    if any(results):
        # filter out null values to avoid pydantic validation errors,
        # for cases where print_df is True and processing errors occur
        report = Report([res for res in results if res])
        report.display_report(config.verbose)


def get_statement_paths(files: Iterable[Path]) -> set[Path]:
    """Recursively collects paths to PDF files from a given collection of paths."""
    matched_files = set()
    for path in files:
        if path.is_file() and str(path).endswith(".pdf"):
            matched_files.add(path)

        if path.is_dir():
            matched_files |= get_statement_paths(path.iterdir())

    return matched_files


@click.command()
@click.version_option(package_name="monopoly-core")
@click.argument(
    "files",
    nargs=-1,
    type=click.Path(exists=True, allow_dash=True, resolve_path=True, path_type=Path),
)
@click.option(
    "-o",
    "--output",
    "output_directory",
    type=click.Path(exists=True, allow_dash=True, resolve_path=True, path_type=Path),
    help="Specify output folder.",
)
@click.option(
    "--preserve-filename",
    is_flag=True,
    help="Keep the input filename (with .csv extension) instead of generating a new one.",
)
@click.option(
    "-p",
    "--pprint",
    is_flag=True,
    help="Print the processed statement(s).",
)
@click.option(
    "-s",
    "--single-process",
    is_flag=True,
    help=("Runs `monopoly` in single-threaded mode, even when processing multiple files. Useful for debugging."),
)
@click.option(
    "--safe/--nosafe",
    "safety_check",
    default=True,
    help=("Determines whether to run the safety check or not. Runs the safety check by default."),
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help=("Increases logging verbosity."),
)
@click.option(
    "--ocr",
    "use_ocr",
    is_flag=True,
    help=("Apply OCR to extract text from scanned documents."),
)
@click.pass_context
@setup_logs
def monopoly(ctx: click.Context, files: list[Path], **kwargs):
    """
    Monopoly converts your bank statements from PDF to CSV.

    A file or directory can be passed in via the FILES argument
    """
    if files:
        matched_files = get_statement_paths(files)

        if matched_files:
            run(matched_files, RunConfig(**kwargs))

        else:
            click.echo(
                click.style("Could not find .pdf files", fg="yellow", bold=True),
                err=True,
            )
            ctx.exit(1)
    else:
        show_welcome_message()


def show_welcome_message():
    art = r"""
     __  __                               _
    |  \/  | ___  _ __   ___  _ __   ___ | |_   _
    | |\/| |/ _ \| '_ \ / _ \| '_ \ / _ \| | | | |
    | |  | | (_) | | | | (_) | |_) | (_) | | |_| |
    |_|  |_|\___/|_| |_|\___/| .__/ \___/|_|\__, |
                             |_|            |___/
    """

    click.echo(art)
    message = "Monopoly converts your bank statements from PDF to CSV"
    click.echo(message)
    commands = [
        (
            "monopoly .",
            "process all statements nested in current dir",
        ),
        (
            "monopoly path/to/file.pdf",
            "process a specific bank statement",
        ),
        (
            "monopoly . --output <dir>",
            "saves all results to a specific directory",
        ),
        (
            "monopoly . --pprint",
            "prints results instead of saving them",
        ),
        (
            "monopoly . --ocr",
            "applies OCR to extract text from scanned documents",
        ),
        (
            "monopoly --help",
            "show more options and other usage information",
        ),
    ]
    margin = max(len(cmd) for cmd, _ in commands)
    for cmd, desc in commands:
        styled_cmd = click.style(
            text=f"{cmd}{' ' * (margin - len(cmd))}",
            fg="white",
            bg="bright_black",
            bold=True,
        )
        click.echo(message=f"    {styled_cmd} {desc}")
    click.echo()



import logging
from functools import wraps

import pytest

logger = logging.getLogger(__name__)


def skip_if_encrypted(func=None):
    """Helper function to skip tests if files are not unlocked with `git-crypt`"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            with open("tests/integration/banks/.gc_check") as f:
                contents = f.read()
                if "unlocked" not in contents:
                    raise ValueError("missing git-crypt unlock")
        except UnicodeDecodeError as err:
            logger.warning(err)
            pytest.skip("Test requires decrypted files. Please run 'git-crypt unlock <KEYFILE>'")
        return func(*args, **kwargs)

    return wrapper if func else skip_if_encrypted



from collections.abc import Sequence
from dataclasses import dataclass, field
from re import Pattern

from monopoly.constants import EntryType
from monopoly.enums import RegexEnum
from monopoly.identifiers import Identifier


@dataclass
class DateOrder:
    """
    Store for date order argument to date parser.

    Supported `dateparser` DATE_ORDER arguments can be found here:
    https://dateparser.readthedocs.io/en/latest/settings.html#date-order.
    """

    date_order: str

    @property
    def settings(self):
        return {"DATE_ORDER": self.date_order}


@dataclass
class MultilineConfig:
    multiline_descriptions: bool = False
    multiline_polarity: bool = False
    multiline_statement_date: bool = False
    multiline_transaction_date: bool = False
    include_prev_margin: int | None = None
    description_margin: int = 3


# pylint: disable=too-many-instance-attributes
@dataclass(kw_only=True)
class StatementConfig:
    r"""
    Configuration store for statements that are dynamically generated at runtime.

    Base configuration class storing configuration values for debit and
    credit card statements.

    - `transaction_pattern` refers to the regex pattern used to capture transactions,
    where a pattern like:
        "(?P<transaction_date>\\d+/\\d+)\\s*"
        "(?P<description>.*?)\\s*"
        "(?P<amount>[\\d.,]+)$"
    is used to capture a transaction like:
        06/07 URBAN TRANSIT CO. SINGAPORE SG  1.38
    - `transaction_date_order` represents the datetime format that a specific bank uses
    for transactions. For example, "DMY" will parse 01/02/2024 as 1 Feb 2024.
    Defaults to DMY.
    - `statement_date_format` represents the datetime format that a specific bank uses
    to represent a statement date.
    - `multiline_config` determines whether Monopoly tries to concatenate
    transactions that are split across two lines
    - `header_pattern` is a regex pattern that is used to find the 'header' line
    of a statement, and determine if it is a debit or credit card statement.
    - `transaction_bound` will cause transactions that have an amount past a certain
    number of spaces will be ignored. For example, if `transaction_bound` = 32:
        "01 NOV  BALANCE B/F              190.77" (will be ignored)
        "01 NOV  YA KUN KAYA TOAST  12.00       " (will be kept)
    - `transaction_auto_polarity` controls whether transaction amounts are set as negative.
    or positive if they have 'CR' or '+' as a polarity identifier. Enabled by default.
    If enabled, only 'CR' or '+' will make a transaction positive. Disabled by default.
    - `safety_check` controls whether the safety check for banks. Use
    for banks that don't provide total amount (or total debit/credit)
    in the statement. Enabled by default.
    - `filename_fallback_pattern` is an optional regex pattern that extracts the statement
    date from the filename when it cannot be found in the PDF content. The pattern should
    have two capture groups: (1) month abbreviation and (2) year (e.g., r"_([A-Za-z]{3})(\d{4})")
    to match filenames like "eStatement_Nov2025_*.pdf". Disabled by default (None).
    """

    statement_type: EntryType
    transaction_pattern: Pattern[str] | RegexEnum
    statement_date_pattern: Pattern[str] | RegexEnum
    header_pattern: Pattern[str] | RegexEnum
    transaction_date_order: DateOrder = field(default_factory=lambda: DateOrder("DMY"))
    statement_date_order: DateOrder = field(default_factory=lambda: DateOrder("DMY"))
    transaction_date_format: str = ""
    multiline_config: MultilineConfig = field(default_factory=MultilineConfig)
    transaction_bound: int | None = None
    prev_balance_pattern: Pattern[str] | RegexEnum | None = None
    safety_check: bool = True
    transaction_auto_polarity: bool = True
    filename_fallback_pattern: Pattern[str] | None = None


@dataclass
class PdfConfig:
    """
    Stores PDF configuration values for the `PdfParser` class.

    - `page_range`: A slice representing which pages to process. For
    example, a range of (1, -1) will mean that the first and last pages
    are skipped.
    - `page_bbox`: A tuple representing the bounding box range for every
    page. This is used to avoid weirdness like vertical text, and other
    PDF artifacts that may affect parsing.
    - `ocr_identifiers`: Applies OCR on PDFs with a specific metadata identifier.
    - `remove_vertical_text`: Whether to remove vertical text from the PDF. This
    helps to avoid issues with pdftotext's layout mode. For performance reasons,
    this defaults to False.
    """

    page_range: tuple[int | None, int | None] = (None, None)
    page_bbox: tuple[float, float, float, float] | None = None
    ocr_identifiers: Sequence[Sequence[Identifier]] | None = None
    remove_vertical_text: bool = False



from monopoly.statements import Transaction


def test_transaction_handles_comma():
    transaction = Transaction(transaction_date="2099-09-10", description="foo", amount="123,123.12")
    assert transaction.amount == -123123.12


def test_transaction_handles_whitespace():
    transaction = Transaction(
        transaction_date="2099-09-10",
        description="   foo,     bar    ",
        amount="123,123.12",
    )
    assert transaction.description == "foo, bar"



import logging
import re

from monopoly.banks.base import BankBase
from monopoly.config import MultilineConfig, PdfConfig, StatementConfig
from monopoly.constants import (
    ISO8601,
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier, TextIdentifier

logger = logging.getLogger(__name__)


class Ocbc(BankBase):
    name = BankNames.OCBC

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=ISO8601.DD_MM_YYYY,
        header_pattern=re.compile(r"(TRANSACTION DATE.*DESCRIPTION.*AMOUNT)"),
        prev_balance_pattern=StatementBalancePatterns.OCBC,
        transaction_pattern=CreditTransactionPatterns.OCBC,
        transaction_date_format="%d/%m",
    )

    debit = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=re.compile(rf"\s{ISO8601.DD_MMM_YYYY}$"),
        header_pattern=re.compile(r"(Withdrawal.*Deposit.*Balance)"),
        transaction_pattern=DebitTransactionPatterns.OCBC,
        multiline_config=MultilineConfig(multiline_descriptions=True),
        transaction_bound=170,
        transaction_date_format="%d %b",
    )

    pdf_config = PdfConfig(
        remove_vertical_text=True,
    )

    identifiers = [
        [
            MetadataIdentifier(creator="pdfgen", producer="Streamline PDFGen for OCBC Group"),
            TextIdentifier("OCBC"),
        ],
    ]

    statement_configs = [credit, debit]



from .base import BaseStatement, Transaction
from .credit_statement import CreditStatement
from .debit_statement import DebitStatement

__all__ = ["BaseStatement", "CreditStatement", "DebitStatement", "Transaction"]



import os
from unittest.mock import MagicMock, Mock, patch

import pytest
from pymupdf import Document

from monopoly.banks.detector import BankDetector
from monopoly.config import DateOrder, PdfConfig, StatementConfig
from monopoly.constants import EntryType
from monopoly.handler import StatementHandler
from monopoly.pdf import PdfDocument, PdfPage, PdfParser
from monopoly.statements import BaseStatement, CreditStatement, DebitStatement


@pytest.fixture
def mock_env():
    """Prevents existing environment variables from interfering with tests"""
    with patch.dict(os.environ, clear=True):
        yield


@pytest.fixture
def pdf_document():
    yield PdfDocument(file_path="src/monopoly/examples/example_statement.pdf")


@pytest.fixture
def metadata_analyzer(pdf_document):
    yield BankDetector(pdf_document)


@pytest.fixture
def mock_get_pages():
    with patch.object(PdfParser, "get_pages") as mock_get_pages:
        mock_get_pages.return_value = MagicMock()
        yield mock_get_pages


@pytest.fixture(scope="function")
def handler(parser):
    with patch.object(StatementHandler, "get_statement") as _:
        handler = StatementHandler(parser)
        yield handler


@pytest.fixture(scope="function")
def mock_bank():
    class MockBank:
        pdf_config = PdfConfig()
        passwords = None

    return MockBank()


@pytest.fixture
def parser(mock_bank):
    class MockDocument:
        metadata_identifier = [None]

    document = MockDocument()
    parser = PdfParser(bank=mock_bank, document=document)
    yield parser


def setup_statement_fixture(
    statement_cls: BaseStatement,
    statement_config,
):
    mock_page = Mock(spec=PdfPage)
    mock_page.lines = ["foo", "bar"]
    mock_page.raw_text = ["foo\nbar"]
    document = MagicMock(spec=Document)
    document.name = "mock_document.pdf"
    statement = statement_cls(pages=[mock_page], bank_name="example", config=statement_config, header="foo")
    yield statement


@pytest.fixture(scope="function")
def credit_statement(statement_config):
    yield from setup_statement_fixture(CreditStatement, statement_config)


@pytest.fixture(scope="function")
def debit_statement(statement_config):
    yield from setup_statement_fixture(DebitStatement, statement_config)


@pytest.fixture(scope="function")
def statement(statement_config):
    yield from setup_statement_fixture(BaseStatement, statement_config)


@pytest.fixture(scope="session")
def statement_config():
    statement_config = StatementConfig(
        statement_type=EntryType.CREDIT,
        header_pattern="foo",
        transaction_pattern="foo",
        statement_date_pattern="",
        transaction_date_order=DateOrder("DMY"),
        prev_balance_pattern="foo",
    )
    yield statement_config


@pytest.fixture
def no_banks(monkeypatch):
    monkeypatch.setattr("monopoly.banks.banks", [])



import re

from monopoly.banks.base import BankBase
from monopoly.config import MultilineConfig, StatementConfig
from monopoly.constants import (
    ISO8601,
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.constants.date import DateFormats
from monopoly.identifiers import MetadataIdentifier, TextIdentifier


class TDCanadaTrust(BankBase):
    name = BankNames.TDCT

    debit_personal = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=re.compile(rf"- {ISO8601.MMM_DD}\/{DateFormats.YY}"),
        header_pattern=re.compile(r"Description.*Withdrawals.*Deposits.*Date.*Balance"),
        transaction_pattern=DebitTransactionPatterns.TDCT,
        transaction_date_format="%b%d",
        safety_check=False,  # total amounts are *per page*, not overall
        transaction_auto_polarity=True,
    )

    debit_business = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=re.compile(rf"- {ISO8601.MMM_DD}\/{DateFormats.YY}"),
        header_pattern=re.compile(r"DESCRIPTION.*CHEQUE/DEBIT.*DEPOSIT/CREDIT.*DATE.*BALANCE"),
        transaction_pattern=DebitTransactionPatterns.TDCT,
        transaction_date_format="%b%d",
        safety_check=False,  # total amounts are *per page*, not overall
        transaction_auto_polarity=True,
    )

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(rf"STATEMENT PERIOD.*{ISO8601.MMMM_DD_YYYY}"),
        header_pattern=re.compile(r"(TRANSACTION\s+POSTING)"),
        prev_balance_pattern=StatementBalancePatterns.TDCT,
        transaction_pattern=CreditTransactionPatterns.TDCT,
        transaction_date_format="%b %d",
        multiline_config=MultilineConfig(multiline_descriptions=True, multiline_polarity=True),
        transaction_auto_polarity=False,
    )

    identifiers = [
        # DR personal
        [
            MetadataIdentifier(producer="OpenText Output Transformation Engine -"),
            # ACCOUNT ISSUED BY: THE TORONTO\-DOMINION BANK
            TextIdentifier(
                text="A CC\nOU\nNT\nI\nSS\nU\nED\nBY :\nTH\nE\nT\nOR\nO\nNT O-\nD\nOM\nI\nNI\nO\nN\nB\nAN\nK"
            ),
        ],
        # DR business
        [
            MetadataIdentifier(producer="OpenText Output Transformation Engine -"),
            TextIdentifier(text="Accounts issued by: THE TORONTO-DOMINION BANK"),
        ],
        # CR
        [
            MetadataIdentifier(producer="OpenText Output Transformation Engine -"),
            TextIdentifier(text="TDSTM"),
        ],
    ]

    statement_configs = [debit_personal, debit_business, credit]



from monopoly.banks.detector import BankDetector
from monopoly.identifiers import MetadataIdentifier


def test_metadata_identifiers_match_partial(metadata_analyzer: BankDetector):
    metadata_identifier = MetadataIdentifier(format="PDF 1.6", creator="incorrect creator")
    metadata_analyzer.document.metadata_identifier = metadata_identifier

    bank_metadata_identifiers = [
        MetadataIdentifier(format="PDF 1.6", creator="correct creator"),
    ]

    assert not metadata_analyzer.identifiers_match(bank_metadata_identifiers)


def test_metadata_identifiers_match_wrong(metadata_analyzer: BankDetector):
    metadata_identifier = MetadataIdentifier(format="PDF 1.2", producer="foo")
    metadata_analyzer.metadata_identifier = metadata_identifier

    bank_metadata_identifiers = [MetadataIdentifier(format="PDF 1.6", producer="bar")]
    assert not metadata_analyzer.identifiers_match(bank_metadata_identifiers)


def test_metadata_identifiers_match_correct(metadata_analyzer: BankDetector):
    metadata_identifier = MetadataIdentifier(format="PDF 1.6", creator="correct creator")
    metadata_analyzer.metadata_identifier = metadata_identifier

    bank_metadata_identifiers = [
        MetadataIdentifier(format="PDF 1.6", creator="correct creator"),
    ]

    assert metadata_analyzer.identifiers_match(bank_metadata_identifiers)



from datetime import datetime

from monopoly.config import DateOrder
from monopoly.pipeline import Pipeline
from monopoly.statements import BaseStatement, Transaction


def test_transform_yyyy_transaction(statement: BaseStatement):
    raw_transactions = [
        Transaction(
            transaction_date="01/12/2023",
            description="FAIRPRICE FINEST",
            amount="18.49",
        ),
    ]
    statement.transactions = raw_transactions
    statement.statement_date = datetime(2024, 1, 1)
    statement.config.transaction_date_order = DateOrder("DMY")
    transformed_transactions = Pipeline.transform(statement)
    expected_transactions = [
        Transaction(transaction_date="2023-12-01", description="FAIRPRICE FINEST", amount=18.49),
    ]

    assert transformed_transactions == expected_transactions


def test_transform_cross_year(statement: BaseStatement):
    raw_transactions = [
        Transaction(transaction_date="12/01", description="FAIRPRICE FINEST", amount="18.49"),
        Transaction(transaction_date="28/12", description="DA PAOLO GASTRONOMIA", amount="19.69"),
        Transaction(transaction_date="28/11", description="KOPITIAM", amount="5.00"),
    ]
    statement.transactions = raw_transactions
    statement.statement_date = datetime(2024, 1, 1)
    statement.config.transaction_date_order = DateOrder("DMY")
    transformed_transactions = Pipeline.transform(statement)
    expected_transactions = [
        Transaction(transaction_date="2024-01-12", description="FAIRPRICE FINEST", amount=18.49),
        Transaction(
            transaction_date="2023-12-28",
            description="DA PAOLO GASTRONOMIA",
            amount=19.69,
        ),
        Transaction(transaction_date="2023-11-28", description="KOPITIAM", amount=5.00),
    ]

    assert transformed_transactions == expected_transactions


def test_transform_within_year(statement: BaseStatement):
    raw_transactions = [
        Transaction(transaction_date="12/06", description="FAIRPRICE FINEST", amount="18.49"),
        Transaction(transaction_date="12/06", description="DA PAOLO GASTRONOMIA", amount="19.69"),
    ]

    statement.transactions = raw_transactions
    statement.statement_date = datetime(2023, 7, 1)
    statement.config.transaction_date_order = DateOrder("DMY")

    transformed_transactions = Pipeline.transform(statement)

    expected_transactions = [
        Transaction(transaction_date="2023-06-12", description="FAIRPRICE FINEST", amount=18.49),
        Transaction(
            transaction_date="2023-06-12",
            description="DA PAOLO GASTRONOMIA",
            amount=19.69,
        ),
    ]

    assert transformed_transactions == expected_transactions



import logging
from dataclasses import dataclass
from functools import cached_property
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING

import pdftotext
from pydantic import SecretStr
from pydantic_settings import BaseSettings, SettingsConfigDict
from pymupdf import TEXTFLAGS_TEXT, Document, Page

from monopoly.identifiers import MetadataIdentifier

if TYPE_CHECKING:
    from monopoly.banks import BankBase

logger = logging.getLogger(__name__)

MIN_OCR_TEXT_LENGTH = 10


class MissingOCRError(Exception):
    """Error that is raised when PDF does not contain any selectable text."""


class PdfPasswords(BaseSettings):
    """
    Pydantic model that automatically populates variables from a .env file.

    Also populates using environment variable called `PDF_PASSWORDS`.
    e.g. export PDF_PASSWORDS='["password123", "secretpass"]'.
    """

    pdf_passwords: list[SecretStr] = [SecretStr("")]
    model_config = SettingsConfigDict(env_file=".env", extra="allow")


@dataclass
class PdfPage:
    """Contains the raw text of a PDF page, and allows access to the raw text."""

    raw_text: str

    @cached_property
    def lines(self) -> list[str]:
        return self.raw_text.split("\n")


class WrongPasswordError(Exception):
    """Exception raised when an incorrect password is provided."""


class MissingPasswordError(Exception):
    """Exception raised when the document is encrypted, but no password is provided."""


class BadPasswordFormatError(Exception):
    """Exception raised passwords are not provided in a proper format."""


class PdfDocument(Document):
    """Handles logic related to the opening, unlocking, and storage of a PDF document."""

    def __init__(
        self,
        file_path: Path | None = None,
        file_bytes: bytes | BytesIO | None = None,
        passwords: list[SecretStr] | None = None,
    ):
        self.file_path = file_path
        self.file_bytes = file_bytes
        self.passwords = passwords or PdfPasswords().pdf_passwords

        if not any([self.file_path, self.file_bytes]):
            msg = "Either `file_path` or `file_bytes` must be passed"
            raise RuntimeError(msg)

        if self.file_path and self.file_bytes:
            msg = "Only one of `file_path` or `file_bytes` should be passed"
            raise RuntimeError(msg)

        args = {"filename": self.file_path, "stream": self.file_bytes}
        super().__init__(**args)

    @cached_property
    def metadata_identifier(self):
        return MetadataIdentifier(**self.metadata)

    def unlock_document(self):
        """Attempt to unlock the document using the provided passwords."""
        if not self.is_encrypted:
            return self

        if not self.passwords:
            msg = "No password found in PDF configuration"
            raise MissingPasswordError(msg)

        if not any(str(password) for password in self.passwords):
            msg = "Password in PDF configuration is empty"
            raise MissingPasswordError(msg)

        if not isinstance(self.passwords, list):
            msg = "Passwords should be stored in a list"
            raise BadPasswordFormatError(msg)

        if not all(isinstance(item, SecretStr) for item in self.passwords):
            msg = "Passwords should be stored as SecretStr"
            raise BadPasswordFormatError(msg)

        for password in self.passwords:
            if self.authenticate(password.get_secret_value()):
                logger.debug("Successfully authenticated with password")
                return self

        msg = f"Could not open document: {self.name}"
        raise WrongPasswordError(msg)

    @cached_property
    def raw_text(self) -> str:
        """Extracts and returns the text from the PDF."""
        raw_text = ""
        for page in self:
            raw_text += page.get_text()
        return raw_text


class PdfParser:
    def __init__(
        self,
        bank: type["BankBase"],
        document: PdfDocument,
    ):
        """
        Class responsible for parsing PDFs and returning raw text.

        The page_range variable determines which pages are extracted.
        All pages are extracted by default.
        """
        self.bank = bank
        self.document = document
        self.metadata_identifier = document.metadata_identifier

    @property
    def pdf_config(self):
        return self.bank.pdf_config

    @cached_property
    def page_range(self):
        return slice(*self.pdf_config.page_range)

    @cached_property
    def page_bbox(self):
        return self.pdf_config.page_bbox

    @cached_property
    def ocr_available(self) -> bool:
        """Check if the document matches any of the specified OCR identifier groups."""
        if not (identifier_groups := self.pdf_config.ocr_identifiers):
            return False

        return any(all(identifier.matches(self) for identifier in group) for group in identifier_groups)

    @cached_property
    def pages(self):
        return self._get_pages()

    def _get_pages(self) -> list[PdfPage]:
        logger.debug("Extracting text from PDF")
        document = self.document

        num_pages = list(range(document.page_count))
        document.select(num_pages[self.page_range])

        if cropbox := self.page_bbox:
            logger.debug("Will crop pages with crop box %s and remove vertical text", cropbox)

        pages = []
        for page in document:
            if self.page_bbox:
                page.set_cropbox(cropbox)

            if self.pdf_config.remove_vertical_text:
                self._remove_vertical_text(page)

            pages.append(page)

        # certain statements requsire garbage collection, so that duplicate objects
        # do not cause pdftotext to fail due to missing xrefs/null values
        # however, setting `garbage=2` may cause issues with other statements
        # so an initial attempt should be made to run using `garbage=0`
        garbage_values = [0, 2]

        for garbage in garbage_values:
            try:
                pdf_byte_stream = BytesIO(document.tobytes(garbage=garbage))
                pdf = pdftotext.PDF(pdf_byte_stream, physical=True)

                # assume PDF is missing OCR if text is less than 10 chars on every page
                if all(len(page) < MIN_OCR_TEXT_LENGTH for page in pdf):
                    msg = "No selectable text found"
                    raise MissingOCRError(msg)

                return [PdfPage(page) for page in pdf]
            except pdftotext.Error:
                continue
        msg = "Unable to retrieve pages"
        raise RuntimeError(msg)

    @staticmethod
    def _remove_vertical_text(page: Page):
        """
        Remove vertical text, based on writing direction (wdir).

        This helps avoid situations where the PDF is oddly parsed, due to vertical text
        inside the PDF.

        An example of vertical text breaking a transaction:
        ```
        'HEALTHY HARVEST CAFÉ SINGAPORE SG',
        'Co Reg No: 123456',
        '10 NOV 9.80',
        ```

        Note:
        ----
            The 'dir' key represents the tuple (cosine, sine) for the angle.
            If line["dir"] != (1, 0), the text of its spans is rotated.

        """
        for block in page.get_text("dict", flags=TEXTFLAGS_TEXT)["blocks"]:
            for line in block["lines"]:
                writing_direction = line["dir"]
                if writing_direction != (1, 0):
                    page.add_redact_annot(line["bbox"])
        page.apply_redactions(images=0)
        return page

    @staticmethod
    def apply_ocr(document: PdfDocument) -> PdfDocument:
        # pylint: disable=import-outside-toplevel
        try:
            from ocrmypdf import Verbosity, configure_logging, ocr
            from ocrmypdf.exceptions import PriorOcrFoundError, TaggedPDFError
        except ImportError:
            logger.warning("ocrmypdf not installed, skipping OCR")
            return document

        added_ocr = False
        try:
            logger.debug("Attempting to apply OCR on %s", document.name)
            original_metadata = document.metadata
            output_bytes = BytesIO()
            configure_logging(Verbosity.quiet)
            logging.getLogger("ocrmypdf").setLevel(logging.ERROR)
            ocr(
                BytesIO(document.tobytes()),
                output_bytes,
                language="eng",
                tesseract_config="tesseract.cfg",
                progress_bar=False,
                optimize=0,
                fast_web_view=999999,
                output_type="pdf",
            )
            output_bytes.seek(0)
            added_ocr = True
            if added_ocr:
                logger.debug("OCR applied to %s", {document.name})

        except (PriorOcrFoundError, TaggedPDFError) as err:
            logger.debug("OCR skipped: %s", str(err))

        # pylint: disable=attribute-defined-outside-init
        if added_ocr:
            logger.debug("Adding OCR layer to document")
            document = PdfDocument(file_bytes=output_bytes)
            document.metadata = original_metadata
        return document



from datetime import datetime
from hashlib import sha256

from monopoly.statements import BaseStatement


def generate_hash(statement: BaseStatement) -> str:
    """Generate a hash based on PDF metadata."""
    hash_object = sha256()
    hash_object.update(str(statement.transactions).encode("utf-8"))
    return hash_object.hexdigest()[0:6]


def generate_name(
    statement: BaseStatement,
    format_type: str,
    bank_name: str,
    statement_type: str,
    statement_date: datetime,
) -> str:
    """
    Generate a new file name, depending on: bank, account type, statement date, and old file name.

    e.g. 20230720.pdf -> hsbc-credit-2023-06-b960bf1e.csv

    The appended UUID ensures that two statements from the same bank in a
    month will not overwrite each other.
    """
    year = statement_date.year
    month = statement_date.month
    file_uuid = generate_hash(statement)

    filename = f"{bank_name}-{statement_type}-{year}-{month:02d}-{file_uuid}.csv"

    formats = {
        "blob": (
            f"bank_name={bank_name}/"
            f"account_type={statement_type}/"
            f"statement_date={statement_date.isoformat()[:10]}/"
            f"{filename}"
        ),
        "file": filename,
    }

    if format_type in formats:
        return formats[format_type]

    msg = "Invalid format_type"
    raise ValueError(msg)



import re
from dataclasses import dataclass
from unittest.mock import MagicMock, patch

from monopoly.statements import CreditStatement, Transaction


def test_prev_month_balance_found(credit_statement: CreditStatement):
    expected = "LAST MONTH'S BALANCE   84.89"

    pattern = re.compile(r"(?P<description>LAST MONTH'S BALANCE?)\s+(?P<amount>[\d.,]+)")
    credit_statement.config.prev_balance_pattern = pattern
    credit_statement.pages[0].lines = ["LAST MONTH'S BALANCE   84.89"]

    result = credit_statement.get_prev_month_balances()
    assert result[0].group() == expected


def test_multiple_prev_month_balance(credit_statement: CreditStatement):
    expected = [
        "LAST MONTH'S BALANCE   84.89",
        "LAST MONTH'S BALANCE  123.12",
    ]

    pattern = re.compile(r"(?P<description>LAST MONTH'S BALANCE?)\s+(?P<amount>[\d.,]+)")
    credit_statement.config.prev_balance_pattern = pattern
    credit_statement.pages[0].lines = [
        "LAST MONTH'S BALANCE   84.89",
        "foo",
        "LAST MONTH'S BALANCE  123.12",
    ]

    results = [match.group() for match in credit_statement.get_prev_month_balances()]
    assert results == expected


def test_prev_month_balance_not_found(credit_statement: CreditStatement):
    credit_statement.config.prev_balance_pattern = re.compile("foobar")
    credit_statement.pages[0].raw_text = "RawTextNotContainingPreviousMonthBalance"

    result = credit_statement.get_prev_month_balances()
    assert not result


def mock_get_prev_month_balances(transactions):
    @dataclass
    class MockRegexMatch:
        transaction_date: str
        description: str
        amount: int

        def groupdict(self):
            return {
                "transaction_date": self.transaction_date,
                "description": self.description,
                "amount": self.amount,
            }

    return [
        MockRegexMatch(transaction_date="1900-01-01", description="foo", amount=99.99),
        MockRegexMatch(transaction_date="1900-01-01", description="bar", amount=123.12),
    ]


@patch(
    "monopoly.statements.credit_statement.CreditStatement.get_prev_month_balances",
    mock_get_prev_month_balances,
)
def test_inject_prev_month_balance(credit_statement):
    transactions = [MagicMock(spec=Transaction) for _ in range(3)]
    for i in range(3):
        transactions[i].date = "2024-01-01"

    result = credit_statement.post_process_transactions(transactions)
    expected = [
        Transaction(
            transaction_date="2024-01-01",
            description="bar",
            amount=-123.12,
            polarity=None,
        ),
        Transaction(
            transaction_date="2024-01-01",
            description="foo",
            amount=-99.99,
            polarity=None,
        ),
    ]
    assert result[0] in expected
    assert result[1] in expected
    assert len(transactions) == 5



import logging
import re
from functools import cached_property
from typing import ClassVar

from monopoly.banks import BankBase
from monopoly.config import MultilineConfig, PdfConfig, StatementConfig
from monopoly.constants import EntryType, InternalBankNames
from monopoly.handler import StatementHandler
from monopoly.pdf import PdfParser

from .generic import DatePatternAnalyzer

logger = logging.getLogger(__name__)


class GenericBank(BankBase):
    identifiers: ClassVar[list] = []
    statement_configs: ClassVar[list[StatementConfig]] = []
    name = InternalBankNames.GENERIC
    pdf_config = PdfConfig(remove_vertical_text=True)
    """
    Empty bank class with variables that can be populated by
    the `GenericStatementHandler` class
    """


class GenericStatementHandler(StatementHandler):
    def __init__(self, parser: PdfParser):
        bank = parser.bank
        pages = parser.pages
        metadata = parser.metadata_identifier

        self.analyzer = DatePatternAnalyzer(pages, metadata)
        bank.statement_configs = list(filter(None, [self.debit, self.credit]))
        super().__init__(parser)

    # override get_header and ignore passed config, since
    # the header line has already been found
    def get_header(self, _: StatementConfig):
        return self.header_pattern

    @cached_property
    def debit(self):
        if self.statement_type == EntryType.DEBIT:
            logger.debug("Creating debit statement config")

            return StatementConfig(
                statement_type=EntryType.DEBIT,
                transaction_pattern=self.transaction_pattern,
                statement_date_pattern=self.statement_date_pattern,
                multiline_config=MultilineConfig(self.multiline_descriptions),
                header_pattern=re.compile(self.header_pattern),
            )
        return None

    @cached_property
    def credit(self):
        if self.statement_type == EntryType.CREDIT:
            logger.debug("Creating credit statement config")

            return StatementConfig(
                statement_type=EntryType.CREDIT,
                prev_balance_pattern=self.prev_balance_pattern,
                transaction_pattern=self.transaction_pattern,
                statement_date_pattern=self.statement_date_pattern,
                header_pattern=re.compile(self.header_pattern),
                multiline_config=MultilineConfig(self.multiline_descriptions),
            )
        return None

    @cached_property
    def transaction_pattern(self):
        return self.analyzer.create_transaction_pattern()

    @cached_property
    def statement_type(self):
        return self.analyzer.get_statement_type()

    @cached_property
    def statement_date_pattern(self):
        return self.analyzer.create_statement_date_pattern()

    @cached_property
    def multiline_descriptions(self):
        return self.analyzer.check_if_multiline()

    @cached_property
    def header_pattern(self):
        lines = self.analyzer.lines_before_first_transaction
        return self.analyzer.get_debit_statement_header_line(lines)

    @cached_property
    def prev_balance_pattern(self):
        return self.analyzer.create_previous_balance_regex()



from dataclasses import fields
from typing import TYPE_CHECKING

from pydantic.dataclasses import dataclass

if TYPE_CHECKING:
    from monopoly.banks.detector import BankDetector


@dataclass(frozen=True)
class Identifier:
    """Parent class for identifiers, used for type hinting."""

    def matches(self, detector: "BankDetector") -> bool:
        """Check if this identifier is found in the parsed document context."""
        raise NotImplementedError


@dataclass(frozen=True)
class MetadataIdentifier(Identifier):
    """Stores the metadata attributes of a PDF."""

    format: str = ""
    title: str = ""
    author: str = ""
    subject: str = ""
    creator: str = ""
    producer: str = ""

    def matches(self, detector: "BankDetector") -> bool:
        """Check for partial matches against the document's metadata."""
        document_metadata = detector.metadata_identifier
        for field in fields(self):
            # The value from the bank config (e.g. format="PDF 1.6")
            config_value = getattr(self, field.name)
            # The value from the parsed document
            document_value = getattr(document_metadata, field.name)

            # We only check fields that are actually defined in the config
            if not config_value:
                continue

            # The document must have a value to compare against
            if not document_value:
                return False

            # The configured value must be a substring of the document's value
            if config_value not in document_value:
                return False

        return True


@dataclass(frozen=True)
class TextIdentifier(Identifier):
    """Stores a specific string that exists in the content of a PDF."""

    text: str = ""

    def matches(self, detector: "BankDetector") -> bool:
        """Check if the identifier's text exists in the document's raw text."""
        # Ensure the document has raw text before checking
        if not detector.document.raw_text:
            return False
        return self.text in detector.document.raw_text



from datetime import datetime
from pathlib import Path

import pytest
from test_utils.skip import skip_if_encrypted
from test_utils.transactions import get_transactions_as_dict, read_transactions_from_csv

from monopoly.banks import BankBase, Dbs, Maybank, Ocbc
from monopoly.pdf import PdfDocument, PdfParser
from monopoly.pipeline import Pipeline
from monopoly.statements import DebitStatement

test_cases = [
    (Dbs, 2222.68, 1302.88, datetime(2023, 10, 31)),
    (Maybank, 5275.61, 4093.7, datetime(2023, 8, 31)),
    (Ocbc, 6630.79, 5049.55, datetime(2023, 10, 31)),
]


@skip_if_encrypted
@pytest.mark.parametrize(
    "bank, expected_debit_sum, expected_credit_sum, statement_date",
    test_cases,
)
def test_bank_debit_statements(
    bank: BankBase,
    expected_debit_sum: float,
    expected_credit_sum: float,
    statement_date: datetime,
):
    test_directory = Path(__file__).parent / bank.name / "debit"

    document = PdfDocument(test_directory / "input.pdf")
    parser = PdfParser(bank, document)
    pipeline = Pipeline(parser)
    statement: DebitStatement = pipeline.extract()

    # check raw data
    expected_raw_transactions = read_transactions_from_csv(test_directory, "raw.csv")
    raw_transactions_as_dict = get_transactions_as_dict(statement.transactions)

    debit_amounts = [transaction.amount for transaction in statement.transactions if transaction.amount > 0]
    credit_amounts = [transaction.amount for transaction in statement.transactions if transaction.amount < 0]

    debit_sum = round(abs(sum(debit_amounts)), 2)
    credit_sum = round(abs(sum(credit_amounts)), 2)

    assert expected_raw_transactions == raw_transactions_as_dict
    assert debit_sum == expected_debit_sum
    assert credit_sum == expected_credit_sum
    assert statement.statement_date == statement_date
    assert statement.perform_safety_check()

    # check transformed data
    expected_transformed_transactions = read_transactions_from_csv(test_directory, "transformed.csv")
    transformed_transactions = pipeline.transform(statement)
    transformed_transactions_as_dict = get_transactions_as_dict(transformed_transactions)
    assert expected_transformed_transactions == transformed_transactions_as_dict



import pytest
from pymupdf import Document

from monopoly.statements import CreditStatement, DebitStatement
from monopoly.statements.base import SafetyCheckError
from monopoly.statements.transaction import Transaction


def test_credit_safety_check(credit_statement: CreditStatement):
    document = Document()
    page = document.new_page()
    text = "Page 1\n3\nfoo\n02 May\n2.27\n27 Apr\n2.67\ntotal amount 31.50"
    page.lines = text.split("\n")
    page.insert_text(point=(0, 0), text=text)
    credit_statement.pages[0] = page

    credit_statement.document = document

    credit_statement.transactions = [
        Transaction(transaction_date="23/01", description="foo", amount=11.5),
        Transaction(transaction_date="24/01", description="bar", amount=20.0),
    ]

    # the safety check should return True, since the total amount of 31.50
    # is present in the text
    assert credit_statement.perform_safety_check()


def test_debit_safety_check(debit_statement: DebitStatement):
    document = Document()
    page = document.new_page()
    text = "Page 1\n3\nfoo\n02 May\n-2.5\n27 Apr\n2.67\ntotal credit 30.0 total debit 2.5"
    page.lines = text.split("\n")
    page.insert_text(point=(0, 0), text=text)
    debit_statement.pages[0] = page

    debit_statement.document = document

    debit_statement.transactions = [
        Transaction(transaction_date="23/01", description="foo", amount=10.0, polarity="CR"),
        Transaction(transaction_date="24/01", description="bar", amount=20.0, polarity="CR"),
        Transaction(transaction_date="25/01", description="baz", amount=-2.5, polarity="DR"),
    ]

    # the safety check should return True, since the credit sum matches
    # 10 + 20 = 30, and 2.5 is present in the statement
    assert debit_statement.perform_safety_check()


def test_debit_safety_check_failure(debit_statement: DebitStatement):
    document = Document()
    page = document.new_page()
    text = "Page 1\n3\nfoo\n02 May\n-999\n27 Apr\n456\nrandom transaction 123"
    page.lines = text.split("\n")
    page.insert_text(point=(0, 0), text=text)
    debit_statement.pages[0] = page

    debit_statement.document = document
    debit_statement.transactions = [
        Transaction(transaction_date="23/01", description="foo", amount=10.0, polarity="CR"),
        Transaction(transaction_date="24/01", description="bar", amount=20.0, polarity="CR"),
        Transaction(transaction_date="25/01", description="baz", amount=-2.5, polarity="DR"),
    ]

    # the safety check should fail, since the debit sum and credit sum
    # are not present as a number
    with pytest.raises(SafetyCheckError):
        debit_statement.perform_safety_check()


def test_safety_check_does_not_pass_when_total_not_in_document(credit_statement: CreditStatement):
    """
    This test guards against the old logic flaw where perform_safety_check()
    would incorrectly pass as long as the sum of transactions was positive,
    even if the total was not actually in the statement text.
    """

    # Create a dummy document with NO total value mentioned
    document = Document()
    page = document.new_page()
    text = "Statement Page 1\nTransaction A 10.00\nTransaction B 20.00\nNo totals listed here"
    page.lines = text.split("\n")
    page.insert_text(point=(0, 0), text=text)
    credit_statement.pages[0] = page
    credit_statement.document = document

    # Transactions sum to 30.00, but 30.00 is NOT in the document text.
    credit_statement.transactions = [
        Transaction(transaction_date="01/01", description="A", amount=10.0, polarity="CR"),
        Transaction(transaction_date="02/01", description="B", amount=20.0, polarity="CR"),
    ]

    # Old logic would wrongly return True — corrected logic should raise an error.
    with pytest.raises(SafetyCheckError):
        credit_statement.perform_safety_check()



import re
from enum import Enum
from functools import cached_property
from typing import Any

from strenum import StrEnum


# pylint: disable=unused-argument,no-self-argument
class AutoEnum(StrEnum):
    """Generates lower case values for enums e.g. CITIBANK -> citibank."""

    def _generate_next_value_(self: str, *_):  # type: ignore[override]
        return self.lower()


class RegexEnum(Enum):
    """A subclass of `Enum` that converts member to a compiled regular expression pattern."""

    def __new__(cls, pattern):
        """
        Create a new `RegexEnum` member.

        Args:
        ----
            pattern (str): A regular expression pattern.

        """
        if not isinstance(pattern, str):
            msg = f"Pattern must be a string, not {type(pattern)}"
            raise TypeError(msg)

        obj = object.__new__(cls)
        obj._value_ = pattern
        return obj

    @cached_property
    def regex(self):
        return re.compile(self.value)

    def __str__(self):
        return self.value

    def match(self, value) -> re.Match:
        return self.regex.search(value)

    def findall(self, value) -> list[Any]:
        return self.regex.findall(value)

    def finditer(self, value) -> list[re.Match]:
        return self.regex.finditer(value)

    def search(self, value) -> re.Match:
        return self.regex.search(value)



import logging
import re

from monopoly.banks.base import BankBase
from monopoly.config import StatementConfig
from monopoly.constants import (
    ISO8601,
    BankNames,
    CreditTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier, TextIdentifier

logger = logging.getLogger(__name__)


class StandardChartered(BankBase):
    name = BankNames.STANDARD_CHARTERED

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(rf": {ISO8601.DD_MMM_YYYY}$"),
        header_pattern=re.compile(r"(Transaction.*Posting.*Amount)"),
        prev_balance_pattern=StatementBalancePatterns.STANDARD_CHARTERED,
        transaction_pattern=CreditTransactionPatterns.STANDARD_CHARTERED,
        transaction_date_format="%d %b",
    )

    identifiers = [
        [
            MetadataIdentifier(
                title="eStatement",
                producer="iText",
            ),
            TextIdentifier("Standard Chartered"),
        ]
    ]

    statement_configs = [credit]



import logging
import re

from monopoly.banks.base import BankBase
from monopoly.config import DateOrder, MultilineConfig, StatementConfig
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
)
from monopoly.constants.date import ISO8601
from monopoly.identifiers import MetadataIdentifier

logger = logging.getLogger(__name__)


class BankOfAmerica(BankBase):
    name = BankNames.BANK_OF_AMERICA

    debit = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=re.compile(rf"for .* to {ISO8601.MMMM_DD_YYYY}"),
        statement_date_order=DateOrder("MDY"),
        transaction_date_order=DateOrder("MDY"),
        header_pattern=re.compile(r"(Date\s+Description\s+Amount)"),
        transaction_pattern=DebitTransactionPatterns.BANK_OF_AMERICA,
        transaction_date_format="%m/%d/%y",
        multiline_config=MultilineConfig(multiline_descriptions=True),
        safety_check=False,
    )

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(rf"Statement Closing Date\s+{ISO8601.MM_DD_YYYY}"),
        statement_date_order=DateOrder("MDY"),
        transaction_date_order=DateOrder("MDY"),
        header_pattern=re.compile(r"(Date\s+Date\s+Description\s+Number\s+Number\s+Amount\s+Total)"),
        transaction_pattern=CreditTransactionPatterns.BANK_OF_AMERICA,
        transaction_date_format="%m/%d",
        multiline_config=MultilineConfig(multiline_descriptions=True),
        safety_check=False,
    )

    identifiers = [
        [
            MetadataIdentifier(
                format="PDF 1.5",
                creator="Bank of America",
                producer="TargetStream StreamEDS",
            )
        ]
    ]

    statement_configs = [debit, credit]



import re

from monopoly.banks.base import BankBase
from monopoly.config import MultilineConfig, StatementConfig
from monopoly.constants.date import ISO8601, DateFormats
from monopoly.constants.statement import (
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier, TextIdentifier


class RoyalBankOfCanada(BankBase):
    name = BankNames.RBC

    debit_personal = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=re.compile(rf"From {ISO8601.MMMM_DD_YYYY} to (?P<date>{ISO8601.MMMM_DD_YYYY})"),
        header_pattern=re.compile(r"Date\s+Description\s+Withdrawals \(\$\)\s+Deposits \(\$\)\s+Balance\ \(\$\)"),
        transaction_pattern=DebitTransactionPatterns.RBC,
        transaction_date_format="%d %b",
        multiline_config=MultilineConfig(multiline_descriptions=True, multiline_transaction_date=True),
    )

    debit_business = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=re.compile(
            rf"\b({DateFormats.MMMM}\s{DateFormats.D}[,\s]{{1,2}}{DateFormats.YYYY}) to "
            rf"(?P<date>\b({DateFormats.MMMM}\s{DateFormats.D}[,\s]{{1,2}}{DateFormats.YYYY}))"
        ),
        header_pattern=re.compile(
            r"Date\s+Description\s+Cheques\ \&\ Debits\ \(\$\)\s+Deposits\ \&\ Credits\ \(\$\)\s+Balance\ \(\$\)"
        ),
        transaction_pattern=re.compile(
            r"^(?!.*(?:Opening balance|Closing balance))"
            r"(\s+)?"
            rf"(?P<transaction_date>{DateFormats.DD}\s+{DateFormats.MMM})?"  # i.e 08 May
            r"\s+"
            r"(?P<description>.+?)"
            r"\s{2,}"
            r"(?P<amount>\d{1,3}(?:,\d{3})*(?:\.\d{1,2}))"
            r"(?:\s{2,}(?P<balance>-?\s?\d{1,3}(?:,\d{3})*(?:\.\d{1,2})))?"
            r"\s*$"
        ),
        transaction_date_format="%d %b",
        multiline_config=MultilineConfig(multiline_descriptions=True, multiline_transaction_date=True),
    )

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(
            rf"STATEMENT FROM ({ISO8601.MMM_DD}|{ISO8601.MMM_DD_YYYY}) TO "
            rf"(({DateFormats.MMM}\s{DateFormats.D}[,\s]{{1,2}}{DateFormats.YYYY})|{ISO8601.MMM_DD_YYYY})"
        ),
        header_pattern=re.compile(r"(TRANSACTION\s+POSTING)"),
        prev_balance_pattern=StatementBalancePatterns.RBC,
        transaction_pattern=CreditTransactionPatterns.RBC,
        transaction_date_format="%b %d",
    )

    identifiers = [
        # DR personal
        [
            MetadataIdentifier(creator="Symcor Inc.", producer="PDFlib+PDI"),
            TextIdentifier(text="Royal Bank of Canada"),
        ],
        [
            MetadataIdentifier(creator="Symcor Inc.", producer="PDFlib+PDI", title="Statement"),
            TextIdentifier(text="Royal Bank of Canada"),
        ],
        # DR business
        [
            MetadataIdentifier(creator="Symcor Inc.", producer="PDFlib+PDI"),
            TextIdentifier(text="ROYAL BANK OF CANADA"),
        ],
        [
            MetadataIdentifier(creator="Symcor Inc.", producer="PDFlib+PDI", title="Statement"),
            TextIdentifier(text="ROYAL BANK OF CANADA"),
        ],
        # CR
        [
            MetadataIdentifier(creator="Symcor Inc.", producer="PDFlib+PDI"),
            TextIdentifier(text="RBC ROYAL BANK"),
        ],
    ]
    statement_configs = [debit_personal, debit_business, credit]



import re
from dataclasses import dataclass
from datetime import datetime

import pytest

from monopoly.constants import EntryType, SharedPatterns
from monopoly.constants.date import DateFormats
from monopoly.generic import DateMatch, DatePatternAnalyzer
from monopoly.pdf import PdfPage, MetadataIdentifier


@pytest.fixture(scope="session")
def mock_pages():
    pages = [
        PdfPage("                Statement Date   31 Oct 2024"),
        PdfPage("01 Oct   ValueVille   123.12 \n02 Oct   RandomShop 124.50"),
        PdfPage("04 Oct   Fernvale Food   12.12\nPayment Due Date   21 Aug 2024"),
    ]
    yield pages


@pytest.fixture(scope="session")
def mock_metadata():
    yield MetadataIdentifier()


@pytest.fixture(scope="function")
def date_pattern_analyzer(mock_pages, mock_metadata):
    yield DatePatternAnalyzer(pages=mock_pages, metadata=mock_metadata)


@pytest.fixture(scope="session")
def lines_before_first_transaction():
    yield list(
        reversed(
            [
                "DBS Bank Ltd",
                "12 Marina Boulevard, Marina Bay Financial Centre Tower 3, Singapore 018982",
                "www.dbs.com",
                "123123123-1234",
                "                JOHN",
                "                FUBAR",
                "                3 STARLING DRIVE",
                "                #01-01 THOMSON RESIDENCES",
                "                SINGAPORE 123456",
                "                                                       As at 31 Oct 2023",
                "",
                " Details of Your DBS Debit Account                     Account No.: 123-4-56789",
                "",
                " DATE    DETAILS OF TRANSACTIONS     WITHDRAWAL($)    DEPOSIT($)     BALANCE($)",
                "",
                "                          Balance Brought Forward                        222.13",
            ]
        )
    )


@dataclass
class MockDateMatch(DateMatch):
    span: tuple[int, int] = (0, 0)
    line: str = ""
    parsed_date: datetime = datetime(1970, 1, 1)
    raw_date: str = ""
    line_number: int = 0
    page_number: int = 0


@pytest.fixture(scope="session")
def patterns_with_date_matches():
    return {
        "dd_mmm": [
            DateMatch(
                span=(33, 39),
                parsed_date=datetime(2024, 10, 31, 0, 0),
                raw_date="31 Oct",
                page_number=0,
                line_number=0,
                line="                Statement Date   31 Oct 2024",
            ),
            DateMatch(
                span=(0, 6),
                parsed_date=datetime(2024, 10, 1, 0, 0),
                raw_date="01 Oct",
                page_number=1,
                line_number=0,
                line="01 Oct   ValueVille   123.12 ",
            ),
            DateMatch(
                span=(0, 6),
                parsed_date=datetime(2024, 10, 2, 0, 0),
                raw_date="02 Oct",
                page_number=1,
                line_number=1,
                line="02 Oct   RandomShop 124.50",
            ),
            DateMatch(
                span=(0, 6),
                parsed_date=datetime(2024, 10, 4, 0, 0),
                raw_date="04 Oct",
                page_number=2,
                line_number=0,
                line="04 Oct   Fernvale Food   12.12",
            ),
            DateMatch(
                span=(19, 25),
                parsed_date=datetime(2024, 8, 21, 0, 0),
                raw_date="21 Aug",
                page_number=2,
                line_number=1,
                line="Payment Due Date   21 Aug 2024",
            ),
        ],
        "dd_mmm_yyyy": [
            DateMatch(
                span=(33, 44),
                parsed_date=datetime(2024, 10, 31, 0, 0),
                raw_date="31 Oct 2024",
                page_number=0,
                line_number=0,
                line="                Statement Date   31 Oct 2024",
            ),
            DateMatch(
                span=(19, 30),
                parsed_date=datetime(2024, 8, 21, 0, 0),
                raw_date="21 Aug 2024",
                page_number=2,
                line_number=1,
                line="Payment Due Date   21 Aug 2024",
            ),
        ],
        "mmm_dd": [
            DateMatch(
                span=(
                    36,
                    42,
                ),
                parsed_date=datetime(2024, 10, 20, 0, 0),
                raw_date="Oct 20",
                page_number=0,
                line_number=0,
                line="                Statement Date   31 Oct 2024",
            ),
            DateMatch(
                span=(
                    22,
                    28,
                ),
                parsed_date=datetime(2024, 8, 20, 0, 0),
                raw_date="Aug 20",
                page_number=2,
                line_number=1,
                line="Payment Due Date   21 Aug 2024",
            ),
        ],
    }


@pytest.fixture(scope="session")
def matches_with_transaction_posting_dates():
    yield [
        MockDateMatch(
            span=(10, 16),
            raw_date="03 OCT",
            parsed_date=datetime(2024, 10, 3),
            page_number=0,
            line_number=31,
            line="  03 OCT           03 OCT           FAST PAYMENT  1.90",
        ),
        MockDateMatch(
            span=(27, 33),
            raw_date="03 OCT",
            parsed_date=datetime(2024, 10, 3),
            page_number=0,
            line_number=31,
            line="  03 OCT           03 OCT           FAST PAYMENT   1.90",
        ),
        MockDateMatch(
            span=(10, 16),
            raw_date="04 OCT",
            parsed_date=datetime(2024, 10, 4),
            page_number=0,
            line_number=36,
            line="  04 OCT           04 OCT           BILL PAYMENT   123.12",
        ),
        MockDateMatch(
            span=(27, 33),
            raw_date="05 OCT",
            parsed_date=datetime(2024, 10, 5),
            page_number=0,
            line_number=36,
            line="  04 OCT           05 OCT           BILL PAYMENT   123.12",
        ),
    ]


@pytest.fixture(scope="session")
def matches_with_posting_transaction_dates():
    yield [
        MockDateMatch(
            span=(10, 16),
            raw_date="05 OCT",
            parsed_date=datetime(2024, 10, 5),
            page_number=0,
            line_number=36,
            line="  05 OCT           04 OCT           BILL PAYMENT   123.12",
        ),
        MockDateMatch(
            span=(27, 33),
            raw_date="04 OCT",
            parsed_date=datetime(2024, 10, 4),
            page_number=0,
            line_number=36,
            line="  05 OCT           04 OCT           BILL PAYMENT   123.12",
        ),
    ]


def test_get_items_with_same_page_line(date_pattern_analyzer: DatePatternAnalyzer):
    year = datetime.now().year
    expected = [
        ("0-0", datetime(year, 10, 31, 0, 0)),
        ("1-0", datetime(year, 10, 1, 0, 0)),
        ("1-1", datetime(year, 10, 2, 0, 0)),
        ("2-0", datetime(year, 10, 4, 0, 0)),
        ("2-1", datetime(year, 8, 21, 0, 0)),
    ]
    assert date_pattern_analyzer.get_items_with_same_page_and_line() == expected


def test_identify_transaction_date(date_pattern_analyzer: DatePatternAnalyzer, matches_with_transaction_posting_dates):
    date_pattern_analyzer.matches = matches_with_transaction_posting_dates
    result = date_pattern_analyzer.is_transaction_date_first()
    assert result


def test_identify_posting_date(date_pattern_analyzer: DatePatternAnalyzer, matches_with_posting_transaction_dates):
    """Check that in cases where date_1 > date_2, the first date is identified
    as the posting date"""
    date_pattern_analyzer.matches = matches_with_posting_transaction_dates
    result = date_pattern_analyzer.is_transaction_date_first()
    assert not result


def test_create_transaction_pattern_with_transaction_first(
    date_pattern_analyzer: DatePatternAnalyzer,
    matches_with_transaction_posting_dates,
):
    expected = (
        rf"(?P<transaction_date>\b({DateFormats.DD}"
        rf"[\/\-\s.]{DateFormats.MMM}))\s+"
        rf"(?P<posting_date>\b({DateFormats.DD}"
        rf"[\/\-\s.]{DateFormats.MMM}))\s+" + SharedPatterns.DESCRIPTION + SharedPatterns.AMOUNT_EXTENDED
    )

    date_pattern_analyzer.spans = {(10, 16): 2, (27, 33): 2}
    date_pattern_analyzer.matches = matches_with_transaction_posting_dates

    result = date_pattern_analyzer.create_transaction_pattern()
    assert result == re.compile(expected, re.IGNORECASE)


def test_create_transaction_pattern_with_posting_first(
    date_pattern_analyzer: DatePatternAnalyzer,
    matches_with_posting_transaction_dates,
):
    date_pattern_analyzer.spans = {(10, 16): 2, (27, 33): 2}
    date_pattern_analyzer.matches = matches_with_posting_transaction_dates

    expected = (
        rf"(?P<posting_date>\b({DateFormats.DD}"
        rf"[\/\-\s.]{DateFormats.MMM}))\s+"
        rf"(?P<transaction_date>\b({DateFormats.DD}"
        rf"[\/\-\s.]{DateFormats.MMM}))\s+" + SharedPatterns.DESCRIPTION + SharedPatterns.AMOUNT_EXTENDED
    )
    result = date_pattern_analyzer.create_transaction_pattern()
    assert result == re.compile(expected, re.IGNORECASE)


def test_get_statement_type_debit(date_pattern_analyzer: DatePatternAnalyzer):
    debit_lines = [
        MockDateMatch(line="01 OCT   ValueVille    123.12     2,000.00"),
        MockDateMatch(line="02 OCT   ShopMart      50.00      1,950.00"),
        MockDateMatch(line="03 OCT   CafeCafe      5.50       1,944.50"),
    ]
    date_pattern_analyzer.matches = debit_lines
    assert date_pattern_analyzer.get_statement_type() == EntryType.DEBIT


def test_get_statement_type_credit(date_pattern_analyzer: DatePatternAnalyzer):
    credit_lines = [
        MockDateMatch(line="01 OCT   Payment received    1,000.00"),
        MockDateMatch(line="02 OCT   Payment received    1,200.00"),
        MockDateMatch(line="03 OCT   Payment received    1,300.00"),
    ]
    date_pattern_analyzer.matches = credit_lines
    assert date_pattern_analyzer.get_statement_type() == EntryType.CREDIT


def test_get_statement_type_mixed_debit(date_pattern_analyzer: DatePatternAnalyzer):
    mixed_debit_lines = [
        MockDateMatch(line="01 OCT   ValueVille          123.12     2,000.00"),
        MockDateMatch(line="02 OCT   Payment received               1,000.00"),
        MockDateMatch(line="03 OCT   CafeCafe            5.50       1,944.50"),
        MockDateMatch(line="04 OCT   Funds Transfer      20.00      1,924.50"),
    ]
    date_pattern_analyzer.matches = mixed_debit_lines
    assert date_pattern_analyzer.get_statement_type() == EntryType.DEBIT


def test_get_debit_statement_header_line(lines_before_first_transaction, date_pattern_analyzer: DatePatternAnalyzer):
    class MockPdfPage:
        def __init__(self, lines):
            self.lines = lines

    pages = [MockPdfPage(lines_before_first_transaction)]
    date_pattern_analyzer.pages = pages
    date_pattern_analyzer.matches = [MockDateMatch(line_number=22, page_number=0)]
    expected_header = (
        "\\ DATE\\s+DETAILS\\ OF\\ TRANSACTIONS\\s+WITHDRAWAL\\(\\$\\)" + "\\s+DEPOSIT\\(\\$\\)\\s+BALANCE\\(\\$\\)"
    )

    result = date_pattern_analyzer.get_debit_statement_header_line(lines_before_first_transaction)
    assert result == expected_header


def test_check_if_multiline(date_pattern_analyzer: DatePatternAnalyzer):
    mulitline_statement_lines = [
        MockDateMatch(page_number=0, line_number=57),
        MockDateMatch(page_number=0, line_number=61),
        MockDateMatch(page_number=0, line_number=67),
        MockDateMatch(page_number=0, line_number=73),
        MockDateMatch(page_number=0, line_number=79),
        MockDateMatch(page_number=1, line_number=18),
        MockDateMatch(page_number=1, line_number=24),
        MockDateMatch(page_number=1, line_number=28),
        MockDateMatch(page_number=1, line_number=34),
        MockDateMatch(page_number=1, line_number=40),
    ]
    date_pattern_analyzer.matches = mulitline_statement_lines
    assert date_pattern_analyzer.check_if_multiline()


def test_check_if_not_multiline(date_pattern_analyzer: DatePatternAnalyzer):
    single_lines = [
        MockDateMatch(page_number=0, line_number=57),
        MockDateMatch(page_number=0, line_number=58),
        MockDateMatch(page_number=1, line_number=21),
        MockDateMatch(page_number=1, line_number=22),
        MockDateMatch(page_number=1, line_number=23),
    ]
    date_pattern_analyzer.matches = single_lines
    assert not date_pattern_analyzer.check_if_multiline()


def test_create_previous_balance_regex(lines_before_first_transaction, date_pattern_analyzer: DatePatternAnalyzer):
    date_pattern_analyzer.lines_before_first_transaction = lines_before_first_transaction
    expected = rf"(?P<description>Balance Brought Forward)\s+{SharedPatterns.AMOUNT}"
    result = date_pattern_analyzer.create_previous_balance_regex()
    expected == result



import os

import pytest

from monopoly.pdf import PdfPasswords


@pytest.fixture
def create_temporary_env_file(tmp_path):
    env_file = tmp_path / ".env"
    env_content = 'PDF_PASSWORDS=["passwordfoo", "passwordbar"]'
    env_file.write_text(env_content)
    return str(env_file)


def test_load_from_environment_variable():
    os.environ["PDF_PASSWORDS"] = '["password1", "password2"]'
    expected_passwords = ["password1", "password2"]
    passwords = PdfPasswords().pdf_passwords

    assert [pw.get_secret_value() for pw in passwords] == expected_passwords


@pytest.mark.usefixtures("mock_env")
def test_load_from_env_file(create_temporary_env_file):
    env_file = create_temporary_env_file
    passwords = PdfPasswords(_env_file=env_file).pdf_passwords
    expected_passwords = ["passwordfoo", "passwordbar"]
    assert [pw.get_secret_value() for pw in passwords] == expected_passwords



import re

from monopoly.banks.base import BankBase
from monopoly.config import MultilineConfig, StatementConfig
from monopoly.constants.date import ISO8601, DateFormats
from monopoly.constants.statement import (
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    SharedPatterns,
)
from monopoly.identifiers import MetadataIdentifier, TextIdentifier


class Scotiabank(BankBase):
    name = BankNames.SCOTIABANK

    debit_personal = StatementConfig(
        statement_type=EntryType.DEBIT,
        header_pattern=re.compile(r"\s+Date\s+Transactions\s+withdrawn\ \(\$\)\s+deposited\ \(\$\)\s+Balance\ \(\$\)"),
        statement_date_pattern=re.compile(
            rf"Closing Balance on (?P<date>{DateFormats.MMMM}\s+{DateFormats.DD},\s+{DateFormats.YYYY})"
        ),
        transaction_date_format="%b %d",
        transaction_pattern=DebitTransactionPatterns.SCOTIABANK,
        multiline_config=MultilineConfig(multiline_descriptions=True),
    )
    debit_business = StatementConfig(
        statement_type=EntryType.DEBIT,
        header_pattern=re.compile(
            r"(\s+)?Date\s+Description\s+Withdrawals/Debits \(\$\)\s+Deposits/Credits \(\$\)\s+Balance \(\$\)"
        ),
        statement_date_pattern=re.compile(
            r"(?P<last_date>\b(?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s+\d{1,2}\s+\d{4}\b)(?!.*\b(?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s+\d{1,2}\s+\d{4}\b)"
        ),
        transaction_date_format="%m/%d/%Y",
        transaction_pattern=re.compile(
            r"^(?!.*(?:BALANCE FORWARD)).*?"
            rf"(?P<transaction_date>{DateFormats.MM}\/{DateFormats.DD}\/{DateFormats.YYYY})\s+"
            + SharedPatterns.DESCRIPTION
            # cents (float) are optional in bank statement. "\.\d*" -> "(?:\.\d{1,2})?"
            + r"(?P<amount>\(?\d{1,3}(?:,\d{3})*(?:\.\d{1,2})??\)?)\s+"
            + r"(?P<balance>\d{1,3}(?:,\d{3})*(?:\.\d{1,2})?)"
        ),
        multiline_config=MultilineConfig(multiline_descriptions=True),
    )

    credit_variant_1 = StatementConfig(
        statement_type=EntryType.CREDIT,
        header_pattern=re.compile(r"\s+REF\.\#\ DATE\s+DATE\s+DETAILS\s+AMOUNT\(\$\)"),
        prev_balance_pattern=re.compile(
            rf"Previous balance,\s+{ISO8601.MMM_DD}\/"
            r"\d{2,}\s+"
            rf"(?P<balance>${SharedPatterns.COMMA_FORMAT})"
        ),
        statement_date_pattern=re.compile(
            # r"Jun 24, 2025 - Jul 23, 2025"
            rf"- {ISO8601.MMM_DD_YYYY}"
        ),
        transaction_date_format="%b %d",
        transaction_pattern=CreditTransactionPatterns.SCOTIABANK,
        multiline_config=MultilineConfig(multiline_descriptions=True),
    )

    identifiers = [
        # DR personal
        [
            MetadataIdentifier(
                producer="CrawfordTech PDF Driver Version 5.1 64 Bit Build ID 7361 on March 09, 2022 at 20:00:30"
            ),
            TextIdentifier(text="www.scotiabank.com"),
        ],
        # DR business
        [
            MetadataIdentifier(
                creator="BIRT Report Engine 2.5.1 using iText 1.5.4.",
                producer=(
                    "iText 1.5.2 (release for Eclipse/BIRT by lowagie.com); "
                    "modified using iText® 5.5.13.1 ©2000-2019 iText Group NV (AGPL-version)"
                ),
            ),
        ],
        # CR
        [
            MetadataIdentifier(
                producer="CrawfordTech PDF Driver Version 5.1 64 Bit Build ID 7361 on March 09, 2022 at 20:00:30"
            ),
            TextIdentifier(text="\nScotiabank"),
            TextIdentifier(text="Scotia\nCredit\nCard"),
        ],
    ]
    statement_configs = [debit_personal, debit_business, credit_variant_1]



import logging
from typing import Any, ClassVar

from monopoly.config import PdfConfig, StatementConfig

logger = logging.getLogger(__name__)


class BankBase:
    """
    Handle initialization of common variables that are shared between bank processor classes.

    Ensures consistency between bank classes.
    """

    name: ClassVar[str]
    statement_configs: ClassVar[list[StatementConfig]]
    pdf_config: PdfConfig = PdfConfig()
    identifiers: ClassVar[list[list[Any]]]

    def __init_subclass__(cls, **kwargs) -> None:
        if not hasattr(cls, "statement_configs"):
            msg = f"{cls.__class__.__name__} must implement `statement_configs` class variable"
            raise NotImplementedError(msg)
        if not hasattr(cls, "identifiers"):
            msg = f"{cls.__class__.__name__} must implement `identifiers` class variable"
            raise NotImplementedError(msg)
        if not hasattr(cls, "name"):
            msg = f"{cls.__class__.__name__} must implement `name` class variable"
            raise NotImplementedError(msg)

        # validation logic only applies to regular banks
        if cls.identifiers and not any(isinstance(item, list) for item in cls.identifiers):
            msg = "`identifiers` must be a list of lists"
            raise TypeError(msg)

        return super().__init_subclass__(**kwargs)



import logging
import re

from monopoly.constants import EntryType
from monopoly.statements.debit_statement import DebitStatement
from monopoly.statements.transaction import Transaction, TransactionGroupDict, TransactionMatch

from .base import BaseStatement, SafetyCheckError

logger = logging.getLogger(__name__)


class CreditStatement(BaseStatement):
    """A dataclass representation of a credit statement."""

    statement_type = EntryType.CREDIT

    def post_process_transactions(self, transactions) -> list[Transaction]:
        previous_month_balances = self.get_prev_month_balances()
        if previous_month_balances:
            first_transaction_date = next(iter(transactions)).date
            for prev_month_balance in previous_month_balances:
                groupdict = TransactionGroupDict(**prev_month_balance.groupdict())
                groupdict.transaction_date = first_transaction_date
                prev_month_transaction = Transaction(**groupdict)
                transactions.insert(0, prev_month_transaction)
        return transactions

    def pre_process_match(self, transaction_match: TransactionMatch) -> TransactionMatch:
        """Pre-process transactions by adding a debit or credit polarity identifier to the group dict."""
        if (
            self.config.statement_type == EntryType.CREDIT
            and transaction_match.groupdict.polarity
            and transaction_match.groupdict.polarity in ("-")
        ):
            transaction_match.groupdict.polarity = "CR"
        return transaction_match

    def get_prev_month_balances(self) -> list[re.Match]:
        """
        Return the previous month's statement balance as a transaction, if it exists in the statement.

        The date is later replaced with a more accurate date by the statement handler.
        """
        prev_balances: list[re.Match] = []

        if pattern := self.config.prev_balance_pattern:
            for page in self.pages:
                prev_balances.extend(match for line in page.lines if (match := pattern.search(line)))

        return prev_balances

    def perform_safety_check(self) -> bool:
        """
        Check that the total sum of all transactions is present somewhere within the document.

        Text is re-extracted from the page, as some bank-specific bounding-box
        configurations (e.g. HSBC) may preclude the total from being extracted

        Returns `False` if the total does not exist in the document.
        """
        numbers = self.get_all_numbers_from_document()
        transactions = self.transactions

        total_amount = round(sum(t.amount for t in transactions), 2)
        total_amount_found = total_amount in numbers

        # if sum of debit and credit is the same as the total amount
        # then the statement is safe
        if abs(total_amount) in numbers or total_amount_found:
            return True

        # attempt a debit-statement style safety for banks that have
        # debit and credit amounts as separate numbers and not a single total sum
        logger.debug("Running debit statement safety check for credit statement")
        if DebitStatement.perform_safety_check(self):
            return True

        msg = f"Total amount {total_amount} cannot be found in credit statement"
        raise SafetyCheckError(msg)



import re

from monopoly.banks import Hsbc, Ocbc
from monopoly.config import MultilineConfig
from monopoly.pdf import PdfPage
from monopoly.statements import BaseStatement
from monopoly.statements.base import MatchContext
from monopoly.statements.transaction import (
    Transaction,
    TransactionGroupDict,
    TransactionMatch,
)


def test_get_transactions(statement: BaseStatement):
    pattern = Ocbc.credit.transaction_pattern
    statement.config.transaction_pattern = pattern

    statement.pages = [PdfPage("19/06 YA KUN KAYA TOAST 3.20\n20/06 FAIRPRICE FINEST 9.90")]
    transactions = statement.get_transactions()
    expected = [
        Transaction(
            transaction_date="19/06",
            description="YA KUN KAYA TOAST",
            amount=-3.2,
            polarity=None,
        ),
        Transaction(
            transaction_date="20/06",
            description="FAIRPRICE FINEST",
            amount=-9.9,
            polarity=None,
        ),
    ]
    assert transactions == expected


def test_check_bound(statement: BaseStatement):
    pattern = Ocbc.credit.transaction_pattern
    statement.config.transaction_pattern = pattern
    statement.config.transaction_bound = 30

    statement.pages = [PdfPage("19/06 YA KUN KAYA TOAST 3.20\n20/06 FAIRPRICE FINEST             9.90")]
    transactions = statement.get_transactions()
    expected = [
        Transaction(
            transaction_date="19/06",
            description="YA KUN KAYA TOAST",
            amount=-3.2,
            polarity=None,
        ),
    ]
    statement.config.transaction_bound = None
    assert transactions == expected


def test_get_multiline_descriptions(statement: BaseStatement):
    pattern = Hsbc.credit.transaction_pattern
    statement.config.multiline_config = MultilineConfig(True)
    statement.config.transaction_pattern = pattern
    statement.pages = [PdfPage("04 Aug 02 Aug SHOPEE 3.20\n              CCY FEE 1.25\n              SINGAPORE SG")]
    transactions = statement.get_transactions()
    expected = [
        Transaction(
            transaction_date="02 Aug",
            description="SHOPEE CCY FEE 1.25 SINGAPORE SG",
            amount=-3.2,
            polarity=None,
        )
    ]
    assert transactions == expected


def test_process_match_multiline_description(statement: BaseStatement):
    pattern = Hsbc.credit.transaction_pattern
    statement.config.multiline_config = MultilineConfig(True)
    statement.config.transaction_pattern = pattern

    groupdict = {
        "transaction_date": "04 Aug",
        "description": "SHOPEE",
        "amount": "3.20",
        "polarity": None,
    }
    match = TransactionMatch(
        match=re.search("foo", "foo"),
        groupdict=TransactionGroupDict(**groupdict),
        page_number=0,
    )

    # case 1: transaction is found on next line
    line = "04 Aug 02 Aug SHOPEE 3.20"
    lines = ["04 Aug 02 Aug SHOPEE 3.20", "05 Aug 02 Aug ValueVille 3.20"]
    expected_groupdict = {
        "transaction_date": "04 Aug",
        "amount": "3.20",
        "description": "SHOPEE",
        "polarity": None,
    }
    context = MatchContext(line=line, lines=lines, idx=0, description="SHOPEE")
    match = statement.process_match(match, context)
    assert match.groupdict.asdict() == expected_groupdict

    # case 2: description across next line is more than three spaces apart
    line = "04 Aug 02 Aug SHOPEE 3.20"
    lines = ["04 Aug 02 Aug SHOPEE 3.20", "foo", "bar"]
    context = MatchContext(line=line, lines=lines, idx=0, description="SHOPEE")
    assert match.groupdict.asdict() == groupdict



from pathlib import Path

from pytest import raises

from monopoly.pdf import MissingOCRError, PdfDocument, PdfParser

fixture_directory = Path(__file__).parent / "fixtures"


def test_get_pages(parser: PdfParser):
    pdf_document = PdfDocument(file_path="src/monopoly/examples/example_statement.pdf")
    parser.document = pdf_document
    parser.page_range = slice(0, -1)
    assert len(parser._get_pages()) == 3


def test_get_pages_with_no_text(parser: PdfParser):
    pdf_document = PdfDocument(file_path=fixture_directory / "4_pages_blank.pdf")
    parser.document = pdf_document
    parser.page_range = slice(0, -1)

    with raises(MissingOCRError):
        parser._get_pages()


def test_get_pages_invalid_returns_error(parser: PdfParser):
    pdf_document = PdfDocument(file_path=fixture_directory / "4_pages_blank.pdf")
    parser.document = pdf_document
    parser.page_range = slice(99, -99)

    with raises(ValueError, match="bad page number"):
        parser._get_pages()



import logging
import re

from monopoly.banks.base import BankBase
from monopoly.config import MultilineConfig, StatementConfig
from monopoly.constants import BankNames, CreditTransactionPatterns, EntryType
from monopoly.constants.date import ISO8601
from monopoly.identifiers import TextIdentifier

logger = logging.getLogger(__name__)


class Amex(BankBase):
    name = BankNames.AMEX

    platinum = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(rf"From .* to {ISO8601.DD_MM_YYYY}"),
        transaction_date_format="%d.%m.%y",
        header_pattern=re.compile(r"(Details.*Foreign Spending.*Amount)"),
        transaction_pattern=CreditTransactionPatterns.AMEX_PLATINUM,
        multiline_config=MultilineConfig(multiline_polarity=True),
    )
    identifiers = [
        [
            TextIdentifier("The Platinum Credit Card"),
            TextIdentifier("americanexpress.com"),
        ]
    ]

    statement_configs = [platinum]



import logging
import re

from monopoly.banks.base import BankBase
from monopoly.config import PdfConfig, StatementConfig
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier

logger = logging.getLogger(__name__)


class Citibank(BankBase):
    name = BankNames.CITIBANK

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(r"Statement\sDate\s+(.*)"),
        header_pattern=re.compile(r"(DATE.*DESCRIPTION.*AMOUNT)"),
        transaction_date_format="%d %b",
        prev_balance_pattern=StatementBalancePatterns.CITIBANK,
        transaction_pattern=CreditTransactionPatterns.CITIBANK,
        filename_fallback_pattern=re.compile(r"_([A-Za-z]{3})(\d{4})"),
    )

    pdf_config = PdfConfig(
        remove_vertical_text=True,
    )

    identifiers = [
        [
            MetadataIdentifier(
                creator="Ricoh Americas Corporation, AFP2PDF",
                producer="Ricoh Americas Corporation, AFP2PDF",
            )
        ]
    ]

    statement_configs = [credit]



from .generic import DateMatch, DatePatternAnalyzer
from .handler import GenericBank, GenericStatementHandler

__all__ = ["DateMatch", "DatePatternAnalyzer", "GenericBank", "GenericStatementHandler"]



import logging
import re

from monopoly.banks.base import BankBase
from monopoly.config import MultilineConfig, PdfConfig, StatementConfig
from monopoly.constants import (
    ISO8601,
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier, TextIdentifier

logger = logging.getLogger(__name__)


class Maybank(BankBase):
    name = BankNames.MAYBANK

    my_debit = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=re.compile(rf"(?:結單日期)[:\s]+{ISO8601.DD_MM_YY}"),
        header_pattern=re.compile(r"(DATE.*DESCRIPTION.*AMOUNT.*BALANCE)"),
        transaction_date_format="%d/%m/%y",
        transaction_pattern=DebitTransactionPatterns.MAYBANK_MY,
        multiline_config=MultilineConfig(multiline_descriptions=True),
    )

    my_credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=ISO8601.DD_MMM_YY,
        header_pattern=re.compile(r"(Date.*Description.*Amount)"),
        transaction_date_format="%d/%m",
        transaction_pattern=CreditTransactionPatterns.MAYBANK_MY,
        prev_balance_pattern=StatementBalancePatterns.MAYBANK_MY,
        multiline_config=MultilineConfig(multiline_descriptions=True),
    )

    sg_credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(f"AS AT {ISO8601.DD_MMM_YYYY}"),
        header_pattern=re.compile(r"(.*DESCRIPTION OF TRANSACTION.*TRANSACTION AMOUNT)"),
        transaction_pattern=CreditTransactionPatterns.MAYBANK_SG,
        prev_balance_pattern=StatementBalancePatterns.MAYBANK_SG,
    )

    sg_credit_identifier = [[TextIdentifier("maybank2u.com.sg"), TextIdentifier("PAYMENT DUE")]]

    identifiers = [
        *sg_credit_identifier,
        [
            MetadataIdentifier(
                author="Maybank2U.com",
                creator="Maybank2u.com",
                producer="iText",
            ),
        ],
        [
            MetadataIdentifier(
                title="Credit Card Statement",
                author="Maybank2U.com",
                producer="iText",
            ),
        ],
    ]

    pdf_config = PdfConfig(
        ocr_identifiers=sg_credit_identifier,
    )

    statement_configs = [my_credit, my_debit, sg_credit]



from monopoly.config import MultilineConfig, StatementConfig, EntryType
from monopoly.statements.base import BaseStatement
from monopoly.statements.transaction import TransactionGroupDict


def test_carry_forward_date_when_multiline_is_enabled():
    """
    Tests that pre_process_transaction_groupdict correctly carries forward a transaction date
    to subsequent transactions when the feature is enabled in the config.

    This simulates a statement where a date appears once, and all following transactions
    on that page are assumed to share that date until a new one appears.
    """
    multiline_config = MultilineConfig(multiline_transaction_date=True)
    config = StatementConfig(
        multiline_config=multiline_config,
        transaction_pattern=".*",
        statement_date_pattern=".*",
        statement_type=EntryType.CREDIT,
        header_pattern=".*",
        transaction_auto_polarity=True,
    )
    statement = BaseStatement(pages=[], bank_name="Test Bank", config=config, header="")

    # SCENARIO: Process a sequence of transactions
    first_date = "24 FEB 2025"

    # Transaction 1: Has a date. This date should be saved as 'previous_transaction_date'.
    tx_with_date_1 = TransactionGroupDict(
        description="ONLINE PAYMENT TO VENDOR X", amount="150.00", transaction_date=first_date
    )
    result_1 = statement.pre_process_transaction_groupdict(tx_with_date_1)
    assert result_1.transaction_date == first_date
    assert statement.previous_transaction_date == first_date, "The first date should be stored immediately"

    # Transaction 2: Has no date. It should inherit the date from the first transaction.
    tx_without_date_1 = TransactionGroupDict(
        description="ANOTHER TRANSACTION X", amount="200.00", transaction_date=None
    )
    result_2 = statement.pre_process_transaction_groupdict(tx_without_date_1)
    assert result_2.transaction_date == first_date, "The second transaction should inherit the first date"

    # Transaction 3: Has a new date. This should update the 'previous_transaction_date'.
    second_date = "25 FEB 2025"
    tx_with_date_2 = TransactionGroupDict(
        description="SOME OTHER TRANSACTION", amount="150.00", transaction_date=second_date
    )
    result_3 = statement.pre_process_transaction_groupdict(tx_with_date_2)
    assert result_3.transaction_date == second_date, "The third transaction should have its new date"
    assert statement.previous_transaction_date == second_date, "The previous date should be updated"

    # Transaction 4: Has no date. It should inherit the new date from the third transaction.
    tx_without_date_2 = TransactionGroupDict(description="FINAL TRANSACTION", amount="120.00", transaction_date=None)
    result_4 = statement.pre_process_transaction_groupdict(tx_without_date_2)
    assert result_4.transaction_date == second_date, "The fourth transaction should inherit the new date"


def test_date_is_not_carried_forward_when_multiline_is_disabled():
    """
    Tests that if multiline_transaction_date is disabled, a missing transaction date
    remains None and is not filled from a previous transaction.
    """
    multiline_config = MultilineConfig(multiline_transaction_date=False)
    config = StatementConfig(
        multiline_config=multiline_config,
        transaction_pattern=".*",
        statement_date_pattern=".*",
        header_pattern=".*",
        statement_type=EntryType.CREDIT,
        transaction_auto_polarity=True,
    )
    statement = BaseStatement(pages=[], bank_name="Test Bank", config=config, header="")

    # SCENARIO: Process two transactions
    # Process one with a date to set the internal state
    statement.pre_process_transaction_groupdict(
        TransactionGroupDict(description="First", amount="100.00", transaction_date="24 FEB 2025")
    )

    # Now process one without a date
    tx_without_date = TransactionGroupDict(description="Second", amount="50.00", transaction_date=None)
    result = statement.pre_process_transaction_groupdict(tx_without_date)

    # 3. ASSERT: The date should remain None because the feature is off
    assert result.transaction_date is None, "Date should not be carried forward when the feature is disabled"



import re

from monopoly.banks import BankBase
from monopoly.config import MultilineConfig, StatementConfig
from monopoly.constants import (
    ISO8601,
    BankNames,
    EntryType,
)
from monopoly.constants.statement import CreditTransactionPatterns
from monopoly.identifiers import MetadataIdentifier


class CanadianTire(BankBase):
    name = BankNames.CANADIAN_TIRE

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        header_pattern=re.compile(r"\s+DATE\s+DATE\s+TRANSACTION\ DESCRIPTION\s+AMOUNT\ \(\$\)"),
        statement_date_pattern=re.compile(rf"Statement\s+date\s+(?P<statement_date>{ISO8601.MMMM_DD_YYYY})"),
        transaction_pattern=CreditTransactionPatterns.CANADIAN_TIRE,
        transaction_date_format="%b %d",
        multiline_config=MultilineConfig(multiline_descriptions=True),
    )

    identifiers = [[MetadataIdentifier(author="Canadian Tire Bank / Banque Canadian Tire", producer="PDFlib+PDI")]]

    statement_configs = [credit]



import logging
import re

from monopoly.banks.base import BankBase
from monopoly.config import DateOrder, MultilineConfig, StatementConfig
from monopoly.constants import BankNames, CreditTransactionPatterns, EntryType
from monopoly.identifiers import MetadataIdentifier, TextIdentifier

logger = logging.getLogger(__name__)


class Chase(BankBase):
    name = BankNames.CHASE

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(r"Statement Date:\s+(.*)"),
        statement_date_order=DateOrder("MDY"),
        transaction_date_order=DateOrder("MDY"),
        header_pattern=re.compile(r"(.*Transaction.*Merchant Name .*\$ Amount)"),
        transaction_date_format="%m/%d",
        transaction_pattern=CreditTransactionPatterns.CHASE,
        multiline_config=MultilineConfig(multiline_descriptions=True),
    )

    identifiers = [
        [
            MetadataIdentifier(
                format="PDF 1.7",
                producer="OpenText Output Transformation Engine - 23.4",
            ),
            TextIdentifier("Chase"),
        ]
    ]

    statement_configs = [credit]



import re

from monopoly.banks import BankBase
from monopoly.config import MultilineConfig, StatementConfig
from monopoly.constants import (
    ISO8601,
    BankNames,
    DebitTransactionPatterns,
    EntryType,
)
from monopoly.constants.statement import CreditTransactionPatterns
from monopoly.identifiers import MetadataIdentifier, TextIdentifier

MMM_DOT_DD = r"[A-Z][a-z]{2,3}\.\s+\d{1,2}"
MMM_DOT_DD_YYYY = r"[A-Z][a-z]{2,3}\.\s+\d{1,2},\s+\d{4}"


class BankOfMontreal(BankBase):
    name = BankNames.BMO

    debit = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=re.compile(rf"For the period ending.*{ISO8601.MMMM_DD_YYYY}"),
        header_pattern=re.compile(
            r"Date\s+Description\s+from\ your\ account\ \(\$\)\s+to\ your\ account\ \(\$\)\s+Balance\ \(\$\)"
        ),
        transaction_pattern=DebitTransactionPatterns.BMO,
        transaction_date_format="%b %d",
        multiline_config=MultilineConfig(multiline_descriptions=True, multiline_transaction_date=True),
    )
    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(rf"Statement\s+date\s+(?P<statement_date>{MMM_DOT_DD_YYYY})"),
        header_pattern=re.compile(r"DATE\s+DATE\s+DESCRIPTION\s+AMOUNT\ \(\$\)"),
        transaction_pattern=CreditTransactionPatterns.BMO,
        transaction_date_format="%b. %d",
    )

    identifiers = [
        # DR
        [
            TextIdentifier(text="www.bmo.com"),  # debit
        ],
        # CR personal
        [MetadataIdentifier(title="BMO Personal")],
        # CR business
        [MetadataIdentifier(title="BMO Small Business Statements")],
    ]

    statement_configs = [debit, credit]



from .date import ISO8601
from .statement import (
    BankNames,
    Columns,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    InternalBankNames,
    SharedPatterns,
    StatementBalancePatterns,
)

__all__ = [
    "ISO8601",
    "BankNames",
    "Columns",
    "CreditTransactionPatterns",
    "DebitTransactionPatterns",
    "EntryType",
    "InternalBankNames",
    "SharedPatterns",
    "StatementBalancePatterns",
]



from unittest.mock import Mock, patch

import pytest

from monopoly.banks.detector import BankDetector
from monopoly.identifiers import MetadataIdentifier
from monopoly.pdf import PdfDocument


class MockDocument:
    def __init__(self, is_encrypted, metadata):
        self.is_encrypted = is_encrypted
        self.metadata = metadata

    def xref_get_key(self):
        return None


class MockPdfDocument:
    def __init__(self, is_encrypted: bool, metadata: dict):
        self.is_encrypted = is_encrypted
        self.metadata_identifier = MetadataIdentifier(**metadata)

    def open(self):
        return MockDocument(self.is_encrypted, self.metadata_identifier)


@pytest.fixture
def mock_encrypted_document():
    yield MockPdfDocument(
        is_encrypted=True,
        metadata={
            "title": "foo",
            "author": "",
            "subject": "",
            "creator": "baz",
            "producer": "",
        },
    )


@pytest.fixture
def mock_non_encrypted_document():
    yield MockPdfDocument(
        is_encrypted=False,
        metadata={
            "title": "foo",
            "author": "",
            "subject": "",
            "creator": "baz",
            "producer": "",
        },
    )


def test_metadata_identifier(mock_non_encrypted_document):
    with patch.object(PdfDocument, "unlock_document", new_callable=Mock) as mock_unlock:
        mock_unlock.return_value = mock_non_encrypted_document

        expected_identifier = MetadataIdentifier(
            title="foo",
            creator="baz",
        )

        metadata_analyzer = BankDetector(mock_non_encrypted_document)

        assert isinstance(metadata_analyzer.metadata_identifier, MetadataIdentifier)
        assert metadata_analyzer.metadata_identifier == expected_identifier



import logging
import re

from monopoly.banks.base import BankBase
from monopoly.config import MultilineConfig, StatementConfig
from monopoly.constants import (
    ISO8601,
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier, TextIdentifier

logger = logging.getLogger(__name__)


class Uob(BankBase):
    name = BankNames.UOB

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(rf"Statement Date.*{ISO8601.DD_MMM_YYYY}"),
        header_pattern=re.compile(r"(Description of Transaction.*Transaction Amount)"),
        prev_balance_pattern=StatementBalancePatterns.UOB,
        transaction_pattern=CreditTransactionPatterns.UOB,
        multiline_config=MultilineConfig(multiline_descriptions=True),
        transaction_date_format="%d %b",
    )

    debit = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=re.compile(rf"Period: .* to {ISO8601.DD_MMM_YYYY}"),
        header_pattern=re.compile(r"(Date.*Description.*Withdrawals.*Deposits.*Balance)"),
        transaction_pattern=DebitTransactionPatterns.UOB,
        transaction_bound=170,
        multiline_config=MultilineConfig(multiline_descriptions=True),
        transaction_date_format="%d %b",
    )

    identifiers = [
        [
            MetadataIdentifier(
                format="PDF 1.5",
                creator="Vault Rendering Engine",
                producer="Rendering Engine",
            ),
        ],
        [TextIdentifier("card.centre@uobgroup.com")],
    ]
    statement_configs = [credit, debit]



import logging
import re

from monopoly.constants import EntryType
from monopoly.statements.transaction import TransactionMatch

from .base import BaseStatement, SafetyCheckError

logger = logging.getLogger(__name__)


class DebitStatement(BaseStatement):
    """A dataclass representation of a debit statement."""

    statement_type = EntryType.DEBIT

    def pre_process_match(self, transaction_match: TransactionMatch) -> TransactionMatch:
        """Pre-process transactions by adding a debit or credit polarity identifier to the group dict."""
        page_number = transaction_match.page_number
        withdrawal_pos = self.get_withdrawal_pos(page_number)
        deposit_pos = self.get_deposit_pos(page_number)
        polarity = transaction_match.groupdict.polarity

        # If the transaction doesn't have an explicit polarity, attempt to infer from column positions
        if not polarity and withdrawal_pos and deposit_pos:
            polarity = self.get_debit_polarity(transaction_match, withdrawal_pos, deposit_pos)

        match polarity:
            case "-" | "DR":
                polarity = "DR"
            case "+" | "CR":
                polarity = "CR"
            case None:
                polarity = "CR"  # default to credit
            case _:
                error = f"Unsupported polarity type {polarity}"
                raise RuntimeError(error)

        transaction_match.groupdict.polarity = polarity
        return transaction_match

    def get_debit_polarity(self, transaction_match: TransactionMatch, withdrawal_pos: int, deposit_pos: int) -> str:
        """
        Get the accounting polarity for debit card statements.

        Attempts to identify whether a transaction is a debit
        or credit entry based on the distance from the withdrawal
        or deposit columns.
        """
        amount = transaction_match.groupdict.amount
        line: str = transaction_match.match.string
        start_pos = line.find(amount)
        # assume that numbers are right aligned
        end_pos = start_pos + len(amount) - 1
        withdrawal_diff = abs(end_pos - withdrawal_pos)
        deposit_diff = abs(end_pos - deposit_pos)
        return "CR" if withdrawal_diff > deposit_diff else "DR"

    def get_withdrawal_pos(self, page_number: int) -> int | None:
        common_names = ["withdraw", "debit", r"from\ your\ account"]
        for name in common_names:
            if pos := self.get_column_pos(name, page_number=page_number):
                return pos
        logger.debug("%s column not found in header on page %s", common_names, page_number)
        return False

    def get_deposit_pos(self, page_number: int) -> int | None:
        common_names = ["deposit", "credit", r"to\ your\ account"]
        for name in common_names:
            if pos := self.get_column_pos(name, page_number=page_number):
                return pos
        logger.debug("%s column not found in header on page %s", common_names, page_number)
        return False

    def get_column_pos(self, column_type: str, page_number: int) -> int | None:
        pattern = re.compile(rf"{column_type}[\w()$]*", re.IGNORECASE)
        if match := pattern.search(self.header):
            return self.get_header_pos(match.group(), page_number)
        return None

    def get_header_pos(self, column_name: str, page_number: int) -> int:
        """
        Return position of the 'WITHDRAWAL' or 'DEPOSIT' header for a particular page.

        An assumption is made here that numbers are right aligned, meaning
        that if an amount matches with the end of the withdrawal string position,
        the item is in fact a withdrawal

        e.g.
        ```
        DATE         DESCRIPTION          WITHDRAWAL         DEPOSIT
        15 OCT       bill payment             322.07
        16 OCT       item                                     123.12
        ```
        """
        header_pattern = self.config.header_pattern
        lines = self.pages[page_number].lines
        for line in lines:
            if match := header_pattern.search(line):
                header = match.string.lower()
                header_start_pos = header.find(column_name.lower())
                if header_start_pos == -1:
                    continue
                return header_start_pos + len(column_name)

        logger.debug("Debit header %s cannot be found on page %s", column_name, page_number)
        return -1

    def perform_safety_check(self: BaseStatement) -> bool:
        """Check that debit and credit transaction sums exist as a number within the statement."""
        transactions = self.transactions

        numbers = self.get_all_numbers_from_document()

        # add a zero, for cases where the debit statement
        # either is completely debit or credit transactions
        numbers.update([0])

        debit_amounts = [transaction.amount for transaction in transactions if transaction.amount > 0]
        credit_amounts = [transaction.amount for transaction in transactions if transaction.amount < 0]

        debit_sum = round(abs(sum(debit_amounts)), 2)
        credit_sum = round(abs(sum(credit_amounts)), 2)

        result = all([debit_sum in numbers, credit_sum in numbers])
        if not result:
            raise SafetyCheckError(self.failed_safety_message)

        return result



from datetime import datetime
from unittest.mock import MagicMock, patch

import pytest

from monopoly.write import generate_name


@pytest.fixture
def mock_generate_hash():
    with patch("monopoly.write.generate_hash") as mock:
        mock.return_value = "b960bf1e"
        yield mock


@pytest.mark.usefixtures("mock_generate_hash")
def test_generate_name():
    statement = MagicMock()
    bank_name = "hsbc"
    statement_date = datetime(2023, 6, 15)
    statement_type = "credit"

    expected_filename = "hsbc-credit-2023-06-b960bf1e.csv"
    # Test for format_type="file"
    filename = generate_name(
        statement=statement,
        format_type="file",
        bank_name=bank_name,
        statement_type=statement_type,
        statement_date=statement_date,
    )
    assert filename == expected_filename

    # Test for format_type="blob"
    filename = generate_name(
        statement=statement,
        format_type="blob",
        bank_name=bank_name,
        statement_type=statement_type,
        statement_date=statement_date,
    )
    assert filename == f"bank_name=hsbc/account_type=credit/statement_date=2023-06-15/{expected_filename}"

    # Test for invalid format_type
    with pytest.raises(ValueError, match="Invalid format_type"):
        generate_name(
            statement=statement,
            format_type="invalid_format",
            bank_name=bank_name,
            statement_type=statement_type,
            statement_date=statement_date,
        )



from monopoly.constants import CreditTransactionPatterns
from monopoly.pdf import PdfPage
from monopoly.statements import BaseStatement, Transaction


def test_statement_process_refund(statement: BaseStatement):
    page_content = "08 SEP  AIRBNB * FOO123  456 GB  (343.01)\n14 AUG  AIRBNB * FOO123  456 GB  343.01\n"
    page = PdfPage(raw_text=page_content)
    pattern = CreditTransactionPatterns.CITIBANK
    statement.config.transaction_pattern = pattern
    statement.pages = [page]
    expected_transactions = [
        Transaction(
            transaction_date="08 SEP",
            description="AIRBNB * FOO123 456 GB",
            amount=343.01,
            polarity="CR",
        ),
        Transaction(
            transaction_date="14 AUG",
            description="AIRBNB * FOO123 456 GB",
            amount=-343.01,
            polarity=None,
        ),
    ]
    assert statement.transactions == expected_transactions



from monopoly.banks import ExampleBank
from monopoly.pdf import PdfDocument, PdfParser
from monopoly.pipeline import Pipeline


def example():
    """
    Extract data from example single bank statement.

    You can pass in the bank class if you want to specify a specific bank,
    or use the BankDetector class to try to detect the bank automatically.
    """
    document = PdfDocument(file_path="src/monopoly/examples/example_statement.pdf")
    parser = PdfParser(ExampleBank, document)
    pipeline = Pipeline(parser)

    # This runs pdftotext on the PDF and
    # extracts transactions as raw text
    statement = pipeline.extract()

    # Dates are converted into an ISO 8601 date format
    transactions = pipeline.transform(statement)

    # Parsed transactions writen to a CSV file in the "example" directory
    file_path = pipeline.load(
        transactions=transactions,
        statement=statement,
        output_directory="src/monopoly/examples",
    )

    with open(file_path, encoding="utf8"):
        pass


if __name__ == "__main__":
    example()



from monopoly.statements import BaseStatement


def test_get_decimal_numbers(statement: BaseStatement):
    lines = [
        "These are some sample lines 123.45 and 67.89",
        "Another line with a decimal number 0.56 but not here 123",
        "No decimals in this line 45, maybe here 10.0 or not",
    ]

    expected_decimal_numbers = {123.45, 67.89, 0.56, 10.0}

    decimal_numbers = statement.get_decimal_numbers(lines)

    assert decimal_numbers == expected_decimal_numbers



import logging
from functools import cached_property

from monopoly.config import StatementConfig
from monopoly.constants import EntryType
from monopoly.pdf import PdfParser
from monopoly.statements import BaseStatement, CreditStatement, DebitStatement

logger = logging.getLogger(__name__)


class StatementHandler:
    """
    Retrieve statement information like transactions from the PDF.

    Identifies the statement as either a debit or credit statement based on the debit and credit config.
    """

    def __init__(self, parser: PdfParser):
        self.bank = parser.bank
        self.pages = parser.pages
        self.file_path = parser.document.file_path

    def get_header(self, config: StatementConfig) -> str | None:
        pattern = config.header_pattern

        for page in self.pages:
            for line in page.lines:
                if match := pattern.search(line):
                    return match.group().lower()
        return None

    @cached_property
    def statement(self):
        return self._get_statement()

    def _get_statement(self) -> BaseStatement:
        pages = self.pages
        bank_name = self.bank.name

        for config in self.bank.statement_configs:
            if header := self.get_header(config):
                match config.statement_type:
                    case EntryType.DEBIT:
                        logger.debug("Statement type detected: %s", EntryType.DEBIT)
                        return DebitStatement(pages, bank_name, config, header, self.file_path)
                    case EntryType.CREDIT:
                        logger.debug("Statement type detected: %s", EntryType.CREDIT)
                        return CreditStatement(pages, bank_name, config, header, self.file_path)

        msg = "Could not find header in statement"
        raise RuntimeError(msg)



import pytest
from unittest.mock import Mock, PropertyMock

from monopoly.pdf import PdfParser
from monopoly.identifiers import MetadataIdentifier, TextIdentifier


class MockPdfDocument:
    def __init__(self, metadata: dict, raw_text: str):
        self.metadata_identifier = MetadataIdentifier(**metadata)
        self.raw_text = raw_text


@pytest.fixture
def mock_bank() -> Mock:
    return Mock()


def test_ocr_available_succeeds(mock_bank: Mock):
    mock_document = MockPdfDocument(
        metadata={"creator": "TestCreator", "producer": "TestProducer"}, raw_text="A document containing the MagicWord."
    )
    parser = PdfParser(bank=mock_bank, document=mock_document)

    bank_required_metadata = MetadataIdentifier(creator="TestCreator")
    bank_required_text = TextIdentifier(text="MagicWord")
    identifier_group = [[bank_required_metadata, bank_required_text]]

    type(mock_bank.pdf_config).ocr_identifiers = PropertyMock(return_value=identifier_group)

    assert parser.ocr_available == True


def test_ocr_available_succeeds_with_two_text_identifiers(mock_bank: Mock):
    """
    Tests that ocr_available returns True when a group with two
    TextIdentifiers fully matches the document.
    """
    mock_document = MockPdfDocument(
        metadata={"creator": "AnyCreator"}, raw_text="This document contains keyword_one and also keyword_two."
    )
    parser = PdfParser(bank=mock_bank, document=mock_document)

    bank_required_text_1 = TextIdentifier(text="keyword_one")
    bank_required_text_2 = TextIdentifier(text="keyword_two")
    identifier_group = [[bank_required_text_1, bank_required_text_2]]

    type(mock_bank.pdf_config).ocr_identifiers = PropertyMock(return_value=identifier_group)

    assert parser.ocr_available == True



"""Test filename fallback for statement dates when not found in content."""

import re
from datetime import datetime
from pathlib import Path

import pytest

from monopoly.banks import Citibank
from monopoly.config import StatementConfig
from monopoly.constants import CreditTransactionPatterns, EntryType, StatementBalancePatterns
from monopoly.pdf import PdfPage
from monopoly.statements import CreditStatement


@pytest.mark.parametrize(
    "filename,expected_month",
    [
        ("eStatement_Nov2025_2025-11-07T14_50_26.pdf", 11),
        ("CardStatement_Oct2025.pdf", 10),
    ],
)
def test_filename_fallback_extracts_date(filename, expected_month):
    """Test that statement date can be extracted from filename patterns."""
    pages = [PdfPage(raw_text="DATE  DESCRIPTION  AMOUNT\n15 Nov  TEST TRANSACTION  100.00")]

    file_path = Path(f"/tmp/{filename}")
    statement = CreditStatement(
        pages=pages,
        bank_name="Citibank",
        config=Citibank.credit,
        header="date description amount",
        file_path=file_path,
    )

    result_date = statement.statement_date
    assert result_date.year == 2025
    assert result_date.month == expected_month
    assert result_date.day == 1  # Fallback sets day to 1


def test_content_takes_precedence_over_filename():
    """Test that statement date from content takes precedence over filename."""
    pages_with_date = [
        PdfPage(raw_text="Statement Date 15 Nov 2022\nDATE  DESCRIPTION  AMOUNT\n15 Nov  TEST TRANSACTION  100.00")
    ]

    # Filename suggests December 2025, but content says November 15, 2022
    file_path = Path("/tmp/eStatement_Dec2025_2025-12-07T14_50_26.pdf")
    statement = CreditStatement(
        pages=pages_with_date,
        bank_name="Citibank",
        config=Citibank.credit,
        header="date description amount",
        file_path=file_path,
    )

    result_date = statement.statement_date
    # Should use the date from content, not filename
    assert result_date.year == 2022
    assert result_date.month == 11
    assert result_date.day == 15


def test_filename_fallback_disabled_by_default():
    """Test that filename fallback is disabled when config doesn't have the pattern."""
    pages = [PdfPage(raw_text="DATE  DESCRIPTION  AMOUNT\n15 Nov  TEST TRANSACTION  100.00")]

    # Create a config without filename_fallback_pattern
    config_without_fallback = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(r"Statement\sDate\s+(.*)"),
        header_pattern=re.compile(r"(DATE.*DESCRIPTION.*AMOUNT)"),
        transaction_date_format="%d %b",
        prev_balance_pattern=StatementBalancePatterns.CITIBANK,
        transaction_pattern=CreditTransactionPatterns.CITIBANK,
        # Note: no filename_fallback_pattern set
    )

    file_path = Path("/tmp/eStatement_Nov2025_2025-11-07T14_50_26.pdf")
    statement = CreditStatement(
        pages=pages,
        bank_name="Citibank",
        config=config_without_fallback,
        header="date description amount",
        file_path=file_path,
    )

    # Should raise ValueError since no statement date in content and fallback disabled
    with pytest.raises(ValueError, match="Statement date not found"):
        _ = statement.statement_date



from collections import Counter
from dataclasses import dataclass
from datetime import datetime
from unittest.mock import MagicMock, patch

import pytest

from monopoly.generic.patterns import DateMatch, DatePattern, PatternMatcher
from monopoly.pdf import PdfPage


@pytest.fixture(scope="session")
def mock_pages():
    pages = [
        PdfPage("                Statement Date   31 Oct 2024"),
        PdfPage("01 Oct   ValueVille   123.12 \n02 Oct   RandomShop 124.50"),
        PdfPage("04 Oct   Fernvale Food   12.12\nPayment Due Date   21 Aug 2024"),
    ]
    yield pages


@pytest.fixture(scope="function")
def pattern_matcher(mock_pages):
    yield PatternMatcher(pages=mock_pages)


@pytest.fixture(scope="session")
def lines_before_first_transaction():
    yield list(
        reversed(
            [
                "DBS Bank Ltd",
                "12 Marina Boulevard, Marina Bay Financial Centre Tower 3, Singapore 018982",
                "www.dbs.com",
                "123123123-1234",
                "                JOHN",
                "                FUBAR",
                "                3 STARLING DRIVE",
                "                #01-01 THOMSON RESIDENCES",
                "                SINGAPORE 123456",
                "                                                       As at 31 Oct 2023",
                "",
                " Details of Your DBS Debit Account                     Account No.: 123-4-56789",
                "",
                " DATE    DETAILS OF TRANSACTIONS     WITHDRAWAL($)    DEPOSIT($)     BALANCE($)",
                "",
                "                          Balance Brought Forward                        222.13",
            ]
        )
    )


@dataclass
class MockDateMatch(DateMatch):
    span: tuple[int, int] = (0, 0)
    line: str = ""
    parsed_date: datetime = datetime(1970, 1, 1)
    raw_date: str = ""
    line_number: int = 0
    page_number: int = 0


@pytest.fixture(scope="session")
def lines_with_transaction_posting_dates():
    yield {
        "dd_mmm": [
            MockDateMatch(
                span=(10, 16),
                raw_date="03 OCT",
                parsed_date=datetime(2024, 10, 3),
                page_number=0,
                line_number=31,
                line="  03 OCT           03 OCT           FAST PAYMENT  1.90",
            ),
            MockDateMatch(
                span=(27, 33),
                raw_date="03 OCT",
                parsed_date=datetime(2024, 10, 3),
                page_number=0,
                line_number=31,
                line="  03 OCT           03 OCT           FAST PAYMENT   1.90",
            ),
            MockDateMatch(
                span=(10, 16),
                raw_date="04 OCT",
                parsed_date=datetime(2024, 10, 4),
                page_number=0,
                line_number=36,
                line="  04 OCT           04 OCT           BILL PAYMENT   123.12",
            ),
            MockDateMatch(
                span=(27, 33),
                raw_date="05 OCT",
                parsed_date=datetime(2024, 10, 5),
                page_number=0,
                line_number=36,
                line="  04 OCT           05 OCT           BILL PAYMENT   123.12",
            ),
        ]
    }


@pytest.fixture(scope="session")
def lines_with_posting_transaction_dates():
    yield {
        "dd_mmm": [
            MockDateMatch(
                span=(10, 16),
                raw_date="05 OCT",
                parsed_date=datetime(2024, 10, 5),
                page_number=0,
                line_number=36,
                line="  05 OCT           04 OCT           BILL PAYMENT   123.12",
            ),
            MockDateMatch(
                span=(27, 33),
                raw_date="04 OCT",
                parsed_date=datetime(2024, 10, 4),
                page_number=0,
                line_number=36,
                line="  05 OCT           04 OCT           BILL PAYMENT   123.12",
            ),
        ]
    }


def test_pattern_matcher(pattern_matcher: PatternMatcher):
    pattern: DatePattern = pattern_matcher.get_transaction_pattern()
    spans = pattern_matcher.get_transaction_spans(pattern)

    assert spans == {(0, 6)}
    assert len([pattern for pattern in pattern_matcher]) == 8
    assert pattern.name == "dd_mmm"
    assert pattern.span_occurrences == Counter({(0, 6): 3, (33, 39): 1, (19, 25): 1})
    assert pattern.unique_spans == {(19, 25), (0, 6), (33, 39)}
    assert pattern.matched_lines == [
        "                Statement Date   31 Oct 2024",
        "01 Oct   ValueVille   123.12 ",
        "02 Oct   RandomShop 124.50",
        "04 Oct   Fernvale Food   12.12",
        "Payment Due Date   21 Aug 2024",
    ]


def test_get_transaction_pattern():
    pattern1 = MagicMock(spec=DatePattern)
    pattern1.name = "dd_mmm"
    pattern1.span_occurrences = Counter({(0, 6): 3, (33, 39): 1})

    pattern2 = MagicMock(spec=DatePattern)
    pattern2.name = "dd_mm"
    pattern2.span_occurrences = Counter({(1, 7): 12, (11, 17): 12, (33, 39): 1})

    pattern3 = MagicMock(spec=DatePattern)
    pattern3.name = "mmm_dd"
    pattern3.span_occurrences = Counter({(0, 5): 2})

    with patch.object(PatternMatcher, "set_date_patterns", return_value=None):
        pattern_matcher = PatternMatcher(pages=[])

        pattern_matcher.dd_mmm = pattern1
        pattern_matcher.dd_mmm_yy = pattern2
        pattern_matcher.mmm_dd = pattern3
        pattern = pattern_matcher.get_transaction_pattern()
        spans = pattern_matcher.get_transaction_spans(pattern)
        assert pattern.name == "dd_mm"
        assert spans == {(1, 7), (11, 17)}



from pathlib import Path

import pytest

from monopoly.banks import Dbs, ExampleBank
from monopoly.pdf import PdfDocument, PdfParser
from monopoly.pipeline import Pipeline


@pytest.fixture
def pdf_file_bytes():
    file_path = Path("src/monopoly/examples/example_statement.pdf")
    with file_path.open("rb") as f:
        yield f.read()


def test_pipeline_with_bank():
    file_path = Path("src/monopoly/examples/example_statement.pdf")
    document = PdfDocument(file_path)
    parser = PdfParser(ExampleBank, document)
    pipeline = Pipeline(parser)
    pages = parser._get_pages()
    transactions = pipeline.extract(pages).transactions
    assert len(transactions) == 53
    assert transactions[0].description == "LAST MONTH'S BALANCE"


def test_pipeline_with_bad_bank():
    file_path = Path("src/monopoly/examples/example_statement.pdf")
    document = PdfDocument(file_path)
    parser = PdfParser(Dbs, document)
    pipeline = Pipeline(parser)
    pages = parser._get_pages()

    with pytest.raises(ValueError, match="No transactions found"):
        pipeline.extract(pages)


def test_pipeline_initialization_with_file_path():
    file_path = Path("src/monopoly/examples/example_statement.pdf")
    try:
        document = PdfDocument(file_path)
        parser = PdfParser(ExampleBank, document)
        pipeline = Pipeline(parser)
        pages = parser._get_pages()
        statement = pipeline.extract(pages)
        transactions = pipeline.transform(statement)
        assert len(transactions) == 53
    except RuntimeError as e:
        pytest.fail(f"Pipeline initialization failed with RuntimeError: {e}")


def test_pipeline_bytes_etl(pdf_file_bytes):
    document = PdfDocument(file_bytes=pdf_file_bytes)
    parser = PdfParser(ExampleBank, document)
    pipeline = Pipeline(parser)

    pages = parser._get_pages()
    statement = pipeline.extract(pages)
    transactions = pipeline.transform(statement)
    assert len(transactions) == 53



import re

from monopoly.banks import BankBase
from monopoly.config import StatementConfig
from monopoly.constants import (
    ISO8601,
    BankNames,
    EntryType,
)
from monopoly.constants.statement import CreditTransactionPatterns
from monopoly.identifiers import MetadataIdentifier


class CapitalOneCanada(BankBase):
    name = BankNames.CAPITAL_ONE
    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        header_pattern=re.compile(r"(\s+)?Transaction Date\s+Posting Date\s+Description\s+Amount"),
        statement_date_pattern=re.compile(rf" - (?P<statement_date>{ISO8601.MMM_DD_YYYY})"),
        transaction_pattern=CreditTransactionPatterns.CAPITAL_ONE,
    )
    identifiers = [
        [
            MetadataIdentifier(
                author="Registered to: CAPITAL1", title="Card_Statement_Canada", creator="OpenText Exstream"
            )
        ]
    ]

    statement_configs = [credit]
